"""
Test module for NikeRun reader base module
"""

import os
import pytest
from pandas import DataFrame, Timedelta, TimedeltaIndex, Timestamp
from runpandas import read_nikerun, read_dir_nikerun
from runpandas import types
from runpandas import exceptions

pytestmark = pytest.mark.stable


@pytest.fixture(scope="session")
def temp_dir(tmpdir_factory):
    return tmpdir_factory.mktemp("data")


@pytest.fixture(scope="session")
def valid_nikerun_filename(tmpdir_factory):
    return tmpdir_factory.getbasetemp().join("activity.json")


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


@pytest.fixture
def pandas_activity(dirpath):
    json_file = os.path.join(dirpath, "nikerun", "sample_nikerun.json")
    return read_nikerun(json_file, to_df=True)


@pytest.fixture
def runpandas_activity(dirpath):
    json_file = os.path.join(dirpath, "nikerun", "sample_nikerun.json")
    return read_nikerun(json_file, to_df=False)


def test_read_file_not_exists(valid_nikerun_filename):
    with pytest.raises(IOError):
        read_nikerun(valid_nikerun_filename)


def test_read_file_invalid_nikerun_file(dirpath):
    json_file = os.path.join(dirpath, "gpx", "run.gpx")
    with pytest.raises(exceptions.InvalidFileError):
        read_nikerun(json_file)


def test_read_file_malformed_nikerun_file(dirpath):
    json_file = os.path.join(dirpath, "nikerun", "malformed_metrics.json")
    with pytest.raises(exceptions.InvalidFileError):
        read_nikerun(json_file)


def test_read_file_malformed_lat_lon_nikerun_file(dirpath):
    json_file = os.path.join(dirpath, "nikerun", "malformed_lat_lon.json")
    with pytest.raises(exceptions.InvalidFileError):
        read_nikerun(json_file)


def test_read_file_malformed_lat_nikerun_file(dirpath):
    json_file = os.path.join(dirpath, "nikerun", "malformed_lat.json")
    with pytest.raises(exceptions.InvalidFileError):
        read_nikerun(json_file)


def test_read_file_outofsync_nikerun_file(dirpath):
    json_file = os.path.join(dirpath, "nikerun", "outofsync.json")
    with pytest.raises(ValueError):
        read_nikerun(json_file)


def test_read_file_nikerun_basic_dataframe(dirpath):
    json_file = os.path.join(dirpath, "nikerun", "sample_nikerun.json")
    activity = read_nikerun(json_file, to_df=True)
    assert isinstance(activity, DataFrame)
    assert isinstance(activity.index, TimedeltaIndex)

    assert activity.size == 2828
    included_data = set(
        [
            "latitude",
            "longitude",
            "elevation",
            "calories",
            "heart_rate",
            "nikefuel",
            "steps",
        ]
    )
    assert included_data <= set(activity.columns.to_list())


def test_read_file_nikerun_basic_activity(dirpath):
    json_file = os.path.join(dirpath, "nikerun", "sample_nikerun.json")
    activity = read_nikerun(json_file, to_df=False)
    assert isinstance(activity, types.Activity)
    assert isinstance(activity.index, TimedeltaIndex)
    assert activity.size == 2828
    included_data = set(["lat", "lon", "alt", "hr", "calories", "nikefuel", "steps"])
    assert included_data <= set(activity.columns.to_list())


test_data = [
    (
        pytest.lazy_fixture("pandas_activity"),
        "latitude",
        0,
        -8.04584830816796,
    ),
    (
        pytest.lazy_fixture("pandas_activity"),
        "longitude",
        0,
        -34.895496899134905,
    ),
    (
        pytest.lazy_fixture("runpandas_activity"),
        "lat",
        0,
        -8.04584830816796,
    ),
    (
        pytest.lazy_fixture("runpandas_activity"),
        "lon",
        0,
        -34.895496899134905,
    ),
]


@pytest.mark.parametrize("activity,column,index,expected", test_data)
def test_nikerun_values(activity, column, index, expected):
    assert activity[column].iloc[index] == expected
    assert activity.index[-1].floor(freq="S") == Timedelta("0 days 00:23:08").floor(
        freq="S"
    )

    if isinstance(activity, types.Activity):
        assert activity.start == Timestamp("2020-08-07 09:15:14+00:00")


def test_empty_read_dir(temp_dir):
    activities = read_dir_nikerun(temp_dir)
    assert activities is None


def test_invalid_dir(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    with pytest.raises(AssertionError):
        read_dir_nikerun(tcx_file)


def test_read_dir_full_nikerun(dirpath):
    activities_directory = os.path.join(dirpath, "nikerun", "samples")
    session = read_dir_nikerun(activities_directory)
    assert session.session.count() == 5
    assert isinstance(session, types.Activity)
    size_data = set([2828, 3227, 4515, 6244, 9002])
    assert size_data == set(
        [
            session.xs(index, level=0).size
            for index in session.index.unique(level="start")
        ]
    )


def test_read_file_missing_data_nikerun_basic_activity(dirpath):
    json_file = os.path.join(dirpath, "nikerun", "missing_columns_nikerun.json")
    activity = read_nikerun(json_file, to_df=False)
    assert isinstance(activity, types.Activity)
    assert isinstance(activity.index, TimedeltaIndex)
    assert activity.size == 1212
    included_data = set(["lat", "lon", "alt"])
    assert included_data <= set(activity.columns.to_list())



"""
Test module for TCX reader base module
"""

import os
import pytest
from pandas import DataFrame, Timedelta, TimedeltaIndex, Timestamp
from runpandas import reader
from runpandas import types

pytestmark = pytest.mark.stable


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


@pytest.fixture
def pandas_activity(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "garmin_connect.gpx")
    return reader._read_file(gpx_file, to_df=True)


@pytest.fixture
def runpandas_activity(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "garmin_connect.gpx")
    return reader._read_file(gpx_file, to_df=False)


test_data = [
    (
        pytest.lazy_fixture("pandas_activity"),
        "hr",
        0,
        120,
    ),
    (
        pytest.lazy_fixture("pandas_activity"),
        "hr",
        -1,
        130,
    ),
    (
        pytest.lazy_fixture("pandas_activity"),
        "ele",
        0,
        23.6000003814697265625,
    ),
    (
        pytest.lazy_fixture("pandas_activity"),
        "ele",
        -1,
        23.799999237060546875,
    ),
    (
        pytest.lazy_fixture("pandas_activity"),
        "lat",
        0,
        51.43788929097354412078857421875,
    ),
    (
        pytest.lazy_fixture("pandas_activity"),
        "lon",
        0,
        6.617012657225131988525390625,
    ),
    (
        pytest.lazy_fixture("pandas_activity"),
        "cad",
        -1,
        80,
    ),
    (
        pytest.lazy_fixture("pandas_activity"),
        "cad",
        0,
        70,
    ),
    (
        pytest.lazy_fixture("runpandas_activity"),
        "hr",
        0,
        120,
    ),
    (
        pytest.lazy_fixture("runpandas_activity"),
        "hr",
        -1,
        130,
    ),
    (
        pytest.lazy_fixture("runpandas_activity"),
        "alt",
        0,
        23.6000003814697265625,
    ),
    (
        pytest.lazy_fixture("runpandas_activity"),
        "alt",
        -1,
        23.799999237060546875,
    ),
    (
        pytest.lazy_fixture("runpandas_activity"),
        "lat",
        0,
        51.43788929097354412078857421875,
    ),
    (
        pytest.lazy_fixture("runpandas_activity"),
        "lon",
        0,
        6.617012657225131988525390625,
    ),
    (
        pytest.lazy_fixture("runpandas_activity"),
        "cad",
        -1,
        80,
    ),
    (
        pytest.lazy_fixture("runpandas_activity"),
        "cad",
        0,
        70,
    ),
]


@pytest.mark.parametrize("activity,column,index,expected", test_data)
def test_gpx_values(activity, column, index, expected):
    assert activity[column].iloc[index] == expected
    assert activity.index[-1] == Timedelta("0 days 00:00:02")

    if isinstance(activity, types.Activity):
        assert activity.start == Timestamp("2018-02-21 14:30:50+00:00")


def test_read_file_gpx_basic_dataframe(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "garmin_connect.gpx")
    activity = reader._read_file(gpx_file, to_df=True)
    assert isinstance(activity, DataFrame)
    assert isinstance(activity.index, TimedeltaIndex)

    assert activity.size == 15
    included_data = set(["lat", "lon", "ele", "cad", "hr"])
    assert included_data <= set(activity.columns.to_list())


def test_read_file_gpx_basic_activity(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "garmin_connect.gpx")
    activity = reader._read_file(gpx_file, to_df=False)
    assert isinstance(activity, types.Activity)
    assert isinstance(activity.index, TimedeltaIndex)
    assert activity.size == 15
    included_data = set(["lat", "lon", "alt", "cad", "hr"])
    assert included_data <= set(activity.columns.to_list())


def test_read_file_gpx_with_timestamp_nofrac(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "strava_export.gpx")
    activity = reader._read_file(gpx_file)
    assert activity.start == Timestamp("2021-01-10 08:54:28Z")


def test_read_file_nikerun_gpx_export(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "nike_run.gpx")
    activity = reader._read_file(gpx_file)
    assert activity.start.floor(freq="S") == Timestamp("2021-07-04 11:23:19").floor(
        freq="S"
    )



"""
Tools for parsing Garmin GPX files.
"""
import pandas as pd
from pandas import TimedeltaIndex
from runpandas import _utils as utils
from runpandas import exceptions
from runpandas.types import Activity
from runpandas.types import columns

# According to Garmin, all times are stored in UTC.
DATETIME_FMT = "%Y-%m-%dT%H:%M:%SZ"

# UTC, with fractional seconds
DATETIME_FMT_WITH_FRAC = "%Y-%m-%dT%H:%M:%S.%fZ"


COLUMNS_SCHEMA = {
    "atemp": columns.Temperature,
    "cad": columns.Cadence,
    "ele": columns.Altitude,
    "hr": columns.HeartRate,
    "lon": columns.Longitude,
    "lat": columns.Latitude,
}


def gen_records(file_path):
    nodes = utils.get_nodes(file_path, ("trkpt",), with_root=True)
    root = next(nodes)
    if utils.sans_ns(root.tag) != "gpx":
        raise exceptions.InvalidFileError("gpx")

    trackpoints = nodes
    for trkpt in trackpoints:
        trkpt_dict = utils.recursive_text_extract(trkpt)
        trkpt_dict.update(dict(trkpt.items()))  # lat, lon
        yield trkpt_dict


def read(file_path, to_df=False, **kwargs):
    """
    This method loads a GPX file into a Pandas DataFrame
    or runpandas Activity.
    Column names are translated to runpandas terminology
    (e.g. "HeartRate" > "heart_rate").
    Datetimes indexes are replaced by time offsets.
    All NaN rows are removed.

    Parameters
    ----------
        filename : str, The path to a training file.
        to_df : bool, optional
             Return a obj:`runpandas.Activity` if `to_df=True`,
              otherwise a :obj:`pandas.DataFrame` will be returned.
              Defaults to False.
        **kwargs :
        Keyword args to be passed to the `read` method
              accordingly to the file format.
    Returns
    -------
    Return a obj:`runpandas.Activity` if `to_df=True`, otherwise
             a :obj:`pandas.DataFrame` will be returned.
    """
    data = pd.DataFrame.from_records(gen_records(file_path))
    times = data.pop("time")  # should always be there
    data = data.astype("float64", copy=False)  # try and make numeric
    data.columns = map(utils.camelcase_to_snakecase, data.columns)
    try:
        timestamps = pd.to_datetime(times, format=DATETIME_FMT, utc=True)
    except ValueError:
        # bad format, try with fractional seconds
        try:
            timestamps = pd.to_datetime(times, format=DATETIME_FMT_WITH_FRAC, utc=True)
        except ValueError:
            # bad format with both formats. Trying coercing one after another.
            timestamps = pd.to_datetime(times, format=DATETIME_FMT, errors="coerce")
            mask = timestamps.isnull()
            timestamps.loc[mask] = pd.to_datetime(
                times[mask], format=DATETIME_FMT_WITH_FRAC, errors="coerce"
            )
    timeoffsets = timestamps - timestamps[0]
    timestamp_index = TimedeltaIndex(timeoffsets, unit="s", name="time")
    data.index = timestamp_index
    data.dropna(axis=1, how="all", inplace=True)

    if to_df:
        return data

    return Activity(data, cspecs=COLUMNS_SCHEMA, start=timestamps[0])



from runpandas.types.acessors.moving import _InactivityAssessor  # noqa
from runpandas.types.acessors.metrics import MetricsAcessor  # noqa
from runpandas.types.acessors.session import _SessionAcessor  # noqa



"""
Test module for runpandas acessors
"""

import os
import json
import time
import pytest
from pandas import Timedelta
from stravalib.protocol import ApiV3
from stravalib.client import Client
from stravalib.model import Stream
from runpandas import reader
from runpandas import read_strava
from runpandas import StravaClient

pytestmark = pytest.mark.stable


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


@pytest.fixture(scope="session")
def valid_token_file(tmpdir_factory):
    return tmpdir_factory.getbasetemp().join("token.json")


@pytest.fixture
def strava_client(valid_token_file):
    file_handler = open(valid_token_file, "w")
    file_handler.write(
        '{"access_token": "2334444",  "refresh_token": "235555", "expires_at": %d}'
        % (time.time() + 3600)
    )
    file_handler.close()

    client = StravaClient(
        client_id="STRAVA_ID",
        client_secret="STRAVA_CLIENT_SECRET",
        token_file=valid_token_file,
    )
    return client


class MockResponse:
    def __init__(self, json_file):
        with open(json_file) as json_handler:
            self.json_data = json.load(json_handler)

    def json(self):
        return self.json_data


def mock_get_activity_streams(streams_file):
    """
    @TODO: I needed to mock the behavior the `stravalib.client.get_activity_streams`,
    it isn't the best alternative for mock the request from strava by passing a json file.
    """

    stream_mock = MockResponse(streams_file).json()
    entities = {}
    for key, value in stream_mock.items():
        value["type"] = key
        stream = Stream.deserialize(value)
        entities[stream.type] = stream
    return entities


def test_metrics_validate(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    activity_gpx = reader._read_file(gpx_file, to_df=True)

    # copy of dataframe for testing the index error
    frame_without_index = activity_gpx.copy()
    frame_without_index = frame_without_index.reset_index()
    with pytest.raises(AttributeError):
        frame_without_index.only_moving()

    # dataframe without speed column
    with pytest.raises(AttributeError):
        activity_gpx.only_moving()

    tcx_file = os.path.join(dirpath, "tcx", "stopped_example.tcx")
    activity_tcx = reader._read_file(tcx_file, to_df=True)

    # copy of dataframe for testing the index error
    frame_without_index = activity_tcx.copy()
    frame_without_index = frame_without_index.reset_index()
    with pytest.raises(AttributeError):
        frame_without_index.only_moving()


def test_only_moving_acessor(dirpath, mocker, strava_client):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    frame_gpx = reader._read_file(gpx_file, to_df=False)
    frame_gpx["distpos"] = frame_gpx.compute.distance(correct_distance=False)
    frame_gpx["speed"] = frame_gpx.compute.speed(from_distances=True)
    frame_gpx_only_moving = frame_gpx.only_moving()
    assert frame_gpx_only_moving[~frame_gpx_only_moving["moving"]].shape[0] == 74
    assert frame_gpx_only_moving.ellapsed_time == Timedelta("0 days 01:25:27")
    assert frame_gpx_only_moving.moving_time == Timedelta("0 days 01:14:46")

    tcx_file = os.path.join(dirpath, "tcx", "stopped_example.tcx")
    activity_tcx = reader._read_file(tcx_file, to_df=False)
    frame_tcx_only_moving = activity_tcx.only_moving()
    assert frame_tcx_only_moving[~frame_gpx_only_moving.moving].shape[0] == 74
    assert frame_tcx_only_moving.ellapsed_time == Timedelta("0 days 01:25:27")
    assert frame_tcx_only_moving.moving_time == Timedelta("0 days 01:23:57")

    fit_file = os.path.join(dirpath, "fit", "garmin-fenix-5-basic.fit")
    fit_file = reader._read_file(fit_file, to_df=False)
    frame_fit_only_moving = fit_file.only_moving()
    assert frame_fit_only_moving.ellapsed_time == Timedelta("0 days 00:00:57")
    assert frame_fit_only_moving.moving_time == Timedelta("0 days 00:00:55")

    tcx_file2 = os.path.join(dirpath, "tcx", "basic.tcx")
    frame_tcx_basic = reader._read_file(tcx_file2, to_df=False)
    frame_tcx_basic["distpos"] = frame_tcx_basic.compute.distance(
        correct_distance=False
    )
    frame_tcx_basic["speed"] = frame_tcx_basic.compute.speed(from_distances=True)
    frame_tcx2_only_moving = frame_tcx_basic.only_moving()
    assert frame_tcx2_only_moving.ellapsed_time == Timedelta("0 days 00:33:11")
    assert frame_tcx2_only_moving.moving_time == Timedelta("0 days 00:33:05")

    activity_json = os.path.join(dirpath, "strava", "activity.json")
    streams_json = os.path.join(dirpath, "strava", "streams.json")

    mocker.patch.object(ApiV3, "get", return_value=MockResponse(activity_json).json())

    mocker.patch.object(
        Client,
        "get_activity_streams",
        return_value=mock_get_activity_streams(streams_json),
    )

    # we don't use access token here, since we will mock the stravalib json response
    activity_strava = read_strava(
        activity_id=4437021783,
        access_token=None,
        client=strava_client,
        refresh_token=None,
        to_df=False,
    )

    assert activity_strava.ellapsed_time == Timedelta("0 days 01:25:45")
    assert activity_strava.moving_time == Timedelta("0 days 01:19:41")



"""
Utilities for testing purposes.
"""

import wrapt


def skip_on_exception(exp):
    """
    Skip a test if a specific Exception is raised. This is because
    the Exception is raised for reasons beyond our control (e.g.
    flakey 3rd-party API).
    a signature-preserving decorator
    Parameters
    ----------
    exp : The Exception under which to execute try-except.
    """

    from pytest import skip

    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        try:
            return wrapped(*args, **kwargs)
        except exp as e:
            skip(str(e))

    return wrapper



"""
Test module for reader base module
"""

import os
import pytest
import runpandas
from pandas import DataFrame
from runpandas import reader
from runpandas import exceptions
from runpandas import types

pytestmark = pytest.mark.stable


@pytest.fixture(scope="session")
def invalid_tcx_filename(tmpdir_factory):
    return tmpdir_factory.getbasetemp().join("activity.pwx")


@pytest.fixture(scope="session")
def temp_dir(tmpdir_factory):
    return tmpdir_factory.mktemp("data")


@pytest.fixture(scope="session")
def valid_tcx_filename(tmpdir_factory):
    return tmpdir_factory.getbasetemp().join("activity.tcx")


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


def test_import_module_exists():
    assert reader._import_module("tcx")


def test_top_level_import():
    assert runpandas.read_file == reader._read_file


def test_imort_module_not_exists():
    with pytest.raises(ImportError):
        reader._import_module("json")


def test_read_file_invalid(invalid_tcx_filename):
    invalid_tcx_filename.write("content")
    with pytest.raises(exceptions.InvalidFileError):
        reader._read_file(invalid_tcx_filename)


def test_read_file_not_exists(valid_tcx_filename):
    with pytest.raises(IOError):
        reader._read_file(valid_tcx_filename)


def test_read_file_malformed_tcx(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "malformed.tcx")
    with pytest.raises(exceptions.InvalidFileError):
        reader._read_file(tcx_file)


def test_read_file_tcx_basic_dataframe(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    activity = reader._read_file(tcx_file, to_df=True)
    assert isinstance(activity, DataFrame)
    included_data = set(
        [
            "latitude_degrees",
            "longitude_degrees",
            "altitude_meters",
            "distance_meters",
            "heart_rate_bpm",
        ]
    )
    assert included_data <= set(activity.columns.to_list())
    assert activity.size == 1915


def test_read_file_tcx_basic_activity(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    activity = reader._read_file(tcx_file, to_df=False)
    assert isinstance(activity, types.Activity)
    included_data = set(["lat", "lon", "alt", "dist", "hr"])
    assert included_data <= set(activity.columns.to_list())
    assert activity.size == 1915


def test_measured_series_activity(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    activity = reader._read_file(tcx_file, to_df=False)
    assert isinstance(activity, types.Activity)
    assert isinstance(activity.hr, types.columns.HeartRate)
    assert activity.hr.base_unit == "bpm"
    assert activity.hr.colname == "hr"


def test_read_full_dir(dirpath):
    activities_directory = os.path.join(dirpath, "samples")
    ac_iterator = reader._read_dir(activities_directory, False)
    activities = [ac for ac in ac_iterator]
    assert len(activities) == 8
    assert isinstance(activities[0], types.Activity)
    size_data = set([16289, 12229, 6286, 5656, 6489, 5635, 9485, 9471])
    assert size_data == set([ac.size for ac in activities])


def test_empty_read_dir(temp_dir):
    activities = reader._read_dir(temp_dir)
    assert len([ac for ac in activities]) == 0


def test_invalid_dir(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    with pytest.raises(AssertionError):
        next(reader._read_dir(tcx_file))


def test_read_dir_empty_aggregate(temp_dir):
    session_frame = reader._read_dir_aggregate(temp_dir)
    assert session_frame is None


def test_read_dir_aggregate(dirpath):
    activities_directory = os.path.join(dirpath, "samples")
    session = reader._read_dir_aggregate(activities_directory)
    assert session.session.count() == 8
    assert isinstance(session, types.Activity)
    size_data = set([16289, 12229, 6286, 5656, 6489, 5635, 9485, 9471])
    assert size_data == set(
        [
            session.xs(index, level=0).size
            for index in session.index.unique(level="start")
        ]
    )



"""
Test module for runpandas column types (i.e. MeasureSeries)
"""

import os
import pytest
import pandas as pd
from runpandas import reader
from runpandas.types import Activity, columns

pytestmark = pytest.mark.stable


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


def test_altitude_feet(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    activity_tcx = reader._read_file(tcx_file, to_df=False)
    assert (activity_tcx["alt"].ft[-5]) == 558.6963742234277


def test_altitude_ascent_descent(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    activity_tcx = reader._read_file(tcx_file, to_df=False)
    # test ascent and descent altitudes
    assert (activity_tcx["alt"].ascent[-5]) == 0.4805908200000033
    assert (activity_tcx["alt"].ascent[-1]) == 0.0
    assert (activity_tcx["alt"].descent[-1]) == -1.4420166019999954
    assert (activity_tcx["alt"].descent[-5]) == 0.0


def test_distance_miles(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    activity_gpx = reader._read_file(gpx_file, to_df=False)
    activity_gpx["distpos"] = activity_gpx.compute.distance()
    # test distpos conversion (meters to miles)
    assert (activity_gpx["distpos"].miles[-1]) == 0.0031649143236707027
    # test distance conversion (meters to miles)
    distance = activity_gpx["distpos"].to_distance()
    assert (distance.miles[-1]) == 7.825950111157077


def test_distance_km(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    activity_gpx = reader._read_file(gpx_file, to_df=False)
    activity_gpx["distpos"] = activity_gpx.compute.distance()
    # test distpos conversion (meters to miles)
    assert (activity_gpx["distpos"].km[-1]) == 0.005093437453100809
    # test distance conversion (meters to miles)
    distance = activity_gpx["distpos"].to_distance()
    assert (distance.km[-1]) == 12.594649752172337


def test_speed_kmh(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "stopped_example.tcx")
    activity_tcx = reader._read_file(tcx_file, to_df=False)
    # test conversion method m/s to km/h
    assert (activity_tcx["speed"].kph[-1]) == 8.498772


def test_speed_mph(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "stopped_example.tcx")
    activity_tcx = reader._read_file(tcx_file, to_df=False)
    # test conversion method m/s to miles/h
    assert (activity_tcx["speed"].mph[-1]) == 5.278740372670808


def test_speed_topace(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    activity_gpx = reader._read_file(gpx_file, to_df=False)
    activity_gpx["distpos"] = activity_gpx.compute.distance(correct_distance=True)
    activity_gpx["speed"] = activity_gpx.compute.speed(from_distances=True)
    expected = pd.Timedelta("0 days 00:00:00.392662")
    assert activity_gpx["speed"].to_pace()[-1].total_seconds() == pytest.approx(
        expected.total_seconds()
    )


def test_distpos_distance(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    activity_gpx = reader._read_file(gpx_file, to_df=False)
    activity_gpx["distpos"] = activity_gpx.compute.distance()
    # test distpos to distance
    assert (activity_gpx["distpos"].to_distance()[-1]) == 12594.649752172338


def test_gradient_pct(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    activity_tcx = reader._read_file(tcx_file, to_df=False)
    activity_tcx["grad"] = activity_tcx.compute.gradient()
    assert (activity_tcx["grad"].pct[-1]) == -9.78415717232968


def test_gradient_radians(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    activity_tcx = reader._read_file(tcx_file, to_df=False)
    activity_tcx["grad"] = activity_tcx.compute.gradient()
    assert (activity_tcx["grad"].radians[-1]) == -0.0975311412476771


def test_gradient_degrees(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    activity_tcx = reader._read_file(tcx_file, to_df=False)
    activity_tcx["grad"] = activity_tcx.compute.gradient()
    assert (activity_tcx["grad"].degrees[-1]) == -5.588122764586196


def test_latlon_radians(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    activity_tcx = reader._read_file(tcx_file, to_df=False)
    assert (activity_tcx["lat"].radians[-1]) == pytest.approx(0.6274799778819853, 0.01)
    assert (activity_tcx["lon"].radians[-1]) == -1.3804335163227985


def test_pace_min_km(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "stopped_example.tcx")
    activity_tcx = reader._read_file(tcx_file, to_df=False)
    activity_tcx["speed"] = activity_tcx.compute.speed()
    activity_tcx["pace"] = activity_tcx.compute.pace()
    assert (activity_tcx["pace"].min_per_km[-1]) == pd.Timedelta(
        "0 days 00:07:03.590608"
    )


def test_pace_min_mile(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "stopped_example.tcx")
    activity_tcx = reader._read_file(tcx_file, to_df=False)
    activity_tcx["speed"] = activity_tcx.compute.speed()
    activity_tcx["pace"] = activity_tcx.compute.pace()
    assert (activity_tcx["pace"].min_per_mile[-1].total_seconds()) == pytest.approx(
        pd.Timedelta("0 days 00:04:23.099756").total_seconds()
    )


def test_constructor(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "stopped_example.tcx")
    activity_tcx = reader._read_file(tcx_file, to_df=False)
    activity_tcx["speed"] = activity_tcx.compute.speed()
    # test custom constructor_expanddim series
    speed_frame = activity_tcx["speed"].to_frame()
    assert type(speed_frame) is Activity
    assert isinstance(speed_frame.speed, columns.Speed)
    assert list(speed_frame.columns) == ["speed"]



"""
Test module for summary statistics module
"""
import os
import pytest
import numpy as np
from pandas import Timedelta, Series, concat, isna
from runpandas import reader, read_dir
from pandas.testing import assert_series_equal


pytestmark = pytest.mark.stable


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


@pytest.fixture
def multi_frame(dirpath):
    sessions_dir = os.path.join(dirpath, "samples")
    activities = [activity for activity in read_dir(sessions_dir)]
    keys = [act.start for act in activities]
    multi_frame = concat(activities, keys=keys, names=["start", "time"], axis=0)
    return multi_frame


@pytest.fixture
def simple_activity(dirpath):
    tcx_file = os.path.join(dirpath, "samples", "2020-12-02T06_08_29-300_Running.tcx")
    frame_tcx = reader._read_file(tcx_file, to_df=False)
    return frame_tcx


def test_activity_summary_with_missing_values(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    frame_tcx = reader._read_file(tcx_file, to_df=False)
    result = frame_tcx.summary()

    expected = Series(
        [
            "Running: 26-12-2012 21:29:53",
            4686.31,
            Timedelta("0 days 00:33:11"),
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            156.653,
            np.nan,
            np.nan,
        ],
        index=[
            "Session",
            "Total distance (meters)",
            "Total ellapsed time",
            "Total moving time",
            "Average speed (km/h)",
            "Average moving speed (km/h)",
            "Average pace (per 1 km)",
            "Average pace moving (per 1 km)",
            "Average cadence",
            "Average moving cadence",
            "Average heart rate",
            "Average moving heart rate",
            "Average temperature",
        ],
    )
    assert_series_equal(result, expected)


def test_activity_summary_missing_moving(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    frame_tcx = reader._read_file(tcx_file, to_df=False)
    frame_tcx["distpos"] = frame_tcx.compute.distance(correct_distance=False)
    frame_tcx["speed"] = frame_tcx.compute.speed(from_distances=True)

    # removing hr to simulate the missing values
    frame_tcx.drop("hr", axis=1, inplace=True)

    result = frame_tcx.summary()

    expected = Series(
        [
            "Running: 26-12-2012 21:29:53",
            4686.31,
            Timedelta("0 days 00:33:11"),
            np.nan,
            8.476556294170631,
            np.nan,
            Timedelta("0 days 00:07:04"),
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
        ],
        index=[
            "Session",
            "Total distance (meters)",
            "Total ellapsed time",
            "Total moving time",
            "Average speed (km/h)",
            "Average moving speed (km/h)",
            "Average pace (per 1 km)",
            "Average pace moving (per 1 km)",
            "Average cadence",
            "Average moving cadence",
            "Average heart rate",
            "Average moving heart rate",
            "Average temperature",
        ],
    )
    assert_series_equal(result, expected)


def test_activity_full_summary(dirpath):
    fit_file = os.path.join(dirpath, "fit", "run.fit")
    frame_fit = reader._read_file(fit_file, to_df=False)
    frame_only_moving_fit = frame_fit.only_moving()
    result = frame_only_moving_fit.summary()
    expected = Series(
        [
            "Running: 14-09-2019 15:22:05",
            5839.77,
            Timedelta("0 days 00:52:47"),
            Timedelta("0 days 00:52:21"),
            6.521134575307862,
            6.564336962750716,
            Timedelta("0 days 00:09:12"),
            Timedelta("0 days 00:09:08"),
            63.963394342762065,
            64.0092050209205,
            129.72129783693845,
            129.7623430962343,
            26.905158069883527,
        ],
        index=[
            "Session",
            "Total distance (meters)",
            "Total ellapsed time",
            "Total moving time",
            "Average speed (km/h)",
            "Average moving speed (km/h)",
            "Average pace (per 1 km)",
            "Average pace moving (per 1 km)",
            "Average cadence",
            "Average moving cadence",
            "Average heart rate",
            "Average moving heart rate",
            "Average temperature",
        ],
    )
    assert_series_equal(result, expected)


def test_summary_session(multi_frame, simple_activity):
    multi_frame = multi_frame.session.only_moving()
    summary_frame = multi_frame.session.summarize()
    assert (
        multi_frame.session.count() == summary_frame.shape[0]
    )  # same number of records (activities)

    # check the  first activity summary
    simple_activity_moving = simple_activity.only_moving()
    summary_single_activity = simple_activity_moving.summary()
    # get the same summary from the summary frame
    summary_session_activity = summary_frame.loc[simple_activity.start]
    # compare the results
    assert summary_single_activity.loc["Total distance (meters)"] == pytest.approx(
        summary_session_activity.loc["total_distance"]
    )
    assert summary_single_activity.loc["Total ellapsed time"] == (
        summary_session_activity.loc["ellapsed_time"]
    )
    assert summary_single_activity.loc["Total moving time"] == (
        summary_session_activity.loc["moving_time"]
    )
    assert summary_single_activity.loc["Average speed (km/h)"] == pytest.approx(
        summary_session_activity.loc["mean_speed"] * 3.6
    )
    assert summary_single_activity.loc["Average moving speed (km/h)"] == pytest.approx(
        summary_session_activity.loc["mean_moving_speed"] * 3.6
    )
    assert summary_single_activity.loc["Average pace (per 1 km)"] == (
        summary_session_activity.loc["mean_pace"]
    )
    assert isna(summary_single_activity.loc["Average cadence"])
    assert isna(summary_session_activity.loc["mean_cadence"])
    assert summary_single_activity.loc["Average heart rate"] == pytest.approx(
        summary_session_activity.loc["mean_heart_rate"]
    )
    assert isna(summary_single_activity.loc["Average temperature"])
    assert isna(summary_session_activity.loc["mean_temperature"])



"""
Tools for parsing FIT files.
"""

import pandas as pd
from pandas import TimedeltaIndex
from fitparse import FitFile
from runpandas import _utils as utils
from runpandas.types import Activity
from runpandas.types import columns

COLUMNS_SCHEMA = {
    "distance": columns.Distance,
    "cadence": columns.Cadence,
    "altitude": columns.Altitude,
    "heart_rate": columns.HeartRate,
    "position_long": columns.Longitude._from_semicircles_to_degrees,
    "position_lat": columns.Latitude._from_semicircles_to_degrees,
    "power": columns.Power,
    "speed": columns.Speed,
    "temperature": columns.Temperature,
}


def message_filter(message, keep=("record", "lap", "event")):
    return message.mesg_type is not None and message.mesg_type.name in keep


def gen_records(file_path):
    """Generator function for iterating over *.fit file messages.
    Parameters
    ----------
    file_path : str
        Path to the ANT/Garmin fit file.

    Yields
    ------
        Parsed messages from `file_path`.
    """
    fit_file = FitFile(file_path)

    messages = filter(message_filter, fit_file.get_messages())
    lap = 0
    session = -1

    for record in messages:
        if record.mesg_type.name == "record":
            message = record.get_values()
            message["lap"] = lap
            message["session"] = session
            yield message
        elif record.mesg_type.name == "lap":
            lap += 1
        elif record.mesg_type.name == "event":
            if record.get_value("event_type") == "start":
                # This happens whens an activity is
                # (manually or automatically) paused or
                # stopped and the resumed
                session += 1
        else:
            raise ValueError(
                "Unknown message fit type {0}.".format(record.mesg_type.name)
            )


def read(file_path, to_df=False, **kwargs):
    """
    This method loads a FIT file into a Pandas DataFrame or runpandas Activity.
    Column names are translated to runpandas terminology
    (e.g. "HeartRate" > "heart_rate").
    Datetimes indexes are replaced by time offsets.
    All NaN rows are removed.

    Parameters
    ----------
        filename : str, The path to a training file.
        to_df : bool, optional
             Return a obj:`runpandas.Activity` if `to_df=True`, otherwise
             a :obj:`pandas.DataFrame` will be returned. Defaults to False.
        **kwargs :
        Keyword args to be passed to the `read` method accordingly to the
        file format.
    Returns
    -------
    Return a obj:`runpandas.Activity` if `to_df=True`, otherwise
             a :obj:`pandas.DataFrame` will be returned.
    """
    start = None
    timeoffsets = None

    data = pd.DataFrame.from_records(gen_records(file_path))

    data.columns = map(utils.camelcase_to_snakecase, data.columns)

    if "timestamp" in data:
        timestamps = data.pop("timestamp")
        timestamps = pd.to_datetime(timestamps, utc=True)

        timeoffsets = timestamps - timestamps[0]
        timestamp_index = TimedeltaIndex(timeoffsets, unit="s", name="time")
        data.index = timestamp_index
        start = timestamps[0]

    data.dropna(axis=1, how="all", inplace=True)

    if to_df:
        return data

    return Activity(data, cspecs=COLUMNS_SCHEMA, start=start)



"""Utility functions for loading example datasets"""

import os
import yaml
from typing import List

from urllib.request import urlopen, urlretrieve, Request
from pydantic import parse_obj_as
from runpandas.datasets.schema import ActivityData

ACTIVITIES_INDEX = (
    "https://raw.githubusercontent.com/"
    "corriporai/runpandas-data/master/activities/index.yml"
)


def _get_activity_index(index=ACTIVITIES_INDEX):
    """Report available example activities.
    Requires an internet connection.
    """
    req = Request(index)
    with urlopen(req) as resp:  # nosec
        content = resp.read()
        raw_index = yaml.safe_load(content)

    loaded_data = parse_obj_as(List[ActivityData], raw_index)

    return loaded_data


def _get_config_data(config_path=None):
    if config_path is None:
        config_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "config.yaml"
        )
    with open(config_path, "r") as f:
        config = yaml.safe_load(f.read())
    return config


def _get_cache_path(config=None):
    """Return a path to the cache directory for example datasets.
    This directory is then used by :func:`load_activity`.
    If the ``config`` file is not specified, it tries to read the target dir
    from the ``config.yaml` file.
    """
    op = os.path
    config = _get_config_data(config)
    data_home = op.abspath(op.expanduser(op.expandvars(config["path"]["root"])))

    if not os.path.exists(data_home):
        os.makedirs(data_home)
    return data_home


def activity_examples(path=None, file_type=None, config=None, **kwargs):
    """Load an example activity from the online repository (requires internet).

    This function provides quick access to a small number of example datasets
    that are useful for documenting runpandas or generating reproducible examples
    for tests. It is not necessary for normal usage.

    Use :func:`_get_activity_index` to see a list of available datasets.

    Parameters
    ----------
    name : str, optional
        Name of the dataset (``path`` on
        https://github.com/corriporai/runpandas-data/activities/).
    file_type: str
        Iterates over all files with the given file type and return them.
    config : yaml file, optional
        The directory in which to cache data; see :func:`get_cache_path`.
    kwargs : keys and values, optional
        Additional keyword arguments are passed to passed through to
        :func:`runpandas.read_file`.
    Returns
    -------
    loaded_data : a single or a list of :class:`schema.ActivityData` instances.

    """
    activities = _get_activity_index()
    if path is not None:
        for activity in activities:
            if os.path.basename(activity.path) == path:
                cache_path = os.path.join(
                    _get_cache_path(config), os.path.basename(path)
                )
                if not os.path.exists(cache_path):
                    urlretrieve(activity.path, cache_path)  # nosec
                activity.path = cache_path
                return activity
        raise ValueError(f"'{path}' is not one of the example datasets.")

    if file_type is not None:
        activities = filter(
            lambda file: file.file_type == file_type, activity_examples()
        )

    return activities



"""
Test module for runpandas types i.e Sessions
"""


import os
import pytest
from pandas import Timestamp, concat
from runpandas import read_dir, reader
from runpandas.types import columns
from runpandas.exceptions import RequiredColumnError

pytestmark = pytest.mark.stable


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


@pytest.fixture
def multi_frame(dirpath):
    sessions_dir = os.path.join(dirpath, "samples")
    activities = [activity for activity in read_dir(sessions_dir)]
    keys = [act.start for act in activities]
    multi_frame = concat(activities, keys=keys, names=["start", "time"], axis=0)
    return multi_frame


@pytest.fixture
def simple_activity(dirpath):
    tcx_file = os.path.join(dirpath, "samples", "2020-12-02T06_08_29-300_Running.tcx")
    frame_tcx = reader._read_file(tcx_file, to_df=False)
    return frame_tcx


def test_validate_session(multi_frame, dirpath):
    sessions_dir = os.path.join(dirpath, "samples")
    activities = [activity for activity in read_dir(sessions_dir)]

    assert multi_frame.index.levshape[0] == len(activities)
    assert multi_frame.session.count() == multi_frame.index.levshape[0]

    # test validation error (get the activity)
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")

    with pytest.raises(AssertionError):  # Activity instance
        activity_gpx = reader._read_file(gpx_file, to_df=True)
        activity_gpx.session.count()

    with pytest.raises(AttributeError):  # index errror
        activity_gpx = reader._read_file(gpx_file, to_df=False)
        activity_gpx.session.count()


def test_count_session(multi_frame, dirpath):
    sessions_dir = os.path.join(dirpath, "samples")
    activities = [activity for activity in read_dir(sessions_dir)]

    assert multi_frame.index.levshape[0] == len(activities)
    assert multi_frame.session.count() == multi_frame.index.levshape[0]


def test_distance_session(multi_frame):
    before_count = multi_frame.session.count()
    before_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape
    multi_frame = multi_frame.session.distance()
    assert "distpos" in multi_frame.columns
    assert "dist" in multi_frame.columns
    assert before_count == multi_frame.session.count()  # same number of sessions.

    after_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape
    assert before_shape[0] == after_shape[0]  # same number of records
    assert before_shape[1] + 1 == after_shape[1]  # number of columns + 1 (distpos)

    assert isinstance(multi_frame["distpos"], columns.DistancePerPosition)


def test_speed_session(multi_frame):
    before_count = multi_frame.session.count()
    before_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape

    multi_frame = multi_frame.session.speed()
    assert "speed" in multi_frame.columns
    assert before_count == multi_frame.session.count()  # same number of sessions.

    after_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape
    assert before_shape[0] == after_shape[0]  # same number of records
    assert before_shape[1] == after_shape[1]  # number of columns (speed)

    with pytest.raises(RequiredColumnError):  # index errror
        multi_frame = multi_frame.session.speed(from_distances=True)

    multi_frame = multi_frame.session.distance()
    multi_frame = multi_frame.session.speed(from_distances=True)

    after_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape

    assert "distpos" in multi_frame.columns
    assert "speed" in multi_frame.columns
    assert before_shape[0] == after_shape[0]  # same number of records
    assert before_shape[1] + 1 == after_shape[1]  # number of columns + 1 (distpos)


def test_vertical_speed_session(multi_frame):
    before_count = multi_frame.session.count()
    before_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape

    multi_frame = multi_frame.session.vertical_speed()
    assert "vam" in multi_frame.columns
    assert before_count == multi_frame.session.count()  # same number of sessions.

    after_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape
    assert before_shape[0] == after_shape[0]  # same number of records
    assert before_shape[1] + 1 == after_shape[1]  # number of columns + 1(vam)


def test_gradient_session(multi_frame):
    before_count = multi_frame.session.count()
    before_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape

    multi_frame = multi_frame.session.gradient()
    assert "grad" in multi_frame.columns
    assert before_count == multi_frame.session.count()  # same number of sessions.

    after_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape
    assert before_shape[0] == after_shape[0]  # same number of records
    assert before_shape[1] + 1 == after_shape[1]  # number of columns + 1(grad)


def test_pace_session(multi_frame):
    before_count = multi_frame.session.count()
    before_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape

    multi_frame = multi_frame.session.pace()
    assert "pace" in multi_frame.columns
    assert before_count == multi_frame.session.count()  # same number of sessions.

    after_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape
    assert before_shape[0] == after_shape[0]  # same number of records
    assert before_shape[1] + 1 == after_shape[1]  # number of columns + 1(pace)

    multi_frame = multi_frame.session.distance()
    multi_frame = multi_frame.session.speed(from_distances=True)

    multi_frame = multi_frame.session.pace()

    after_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape

    assert "distpos" in multi_frame.columns
    assert "pace" in multi_frame.columns

    assert before_shape[0] == after_shape[0]  # same number of records
    assert (
        before_shape[1] + 2 == after_shape[1]
    )  # number of columns + 2 (distpos, pace)

    assert isinstance(multi_frame["pace"], columns.Pace)


def test_only_moving_session(multi_frame):
    before_count = multi_frame.session.count()
    before_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape

    multi_frame = multi_frame.session.only_moving()
    assert "moving" in multi_frame.columns
    assert before_count == multi_frame.session.count()  # same number of sessions.

    after_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape
    assert before_shape[0] == after_shape[0]  # same number of records
    assert before_shape[1] + 1 == after_shape[1]  # number of columns + 1(moving)


def test_heart_zone_session(multi_frame):
    before_count = multi_frame.session.count()
    before_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape

    multi_frame = multi_frame.session.heart_zone(
        bins=[0, 92, 110, 129, 147, 166, 184],
        labels=["Rest", "Z1", "Z2", "Z3", "Z4", "Z5"],
    )
    assert "hr_zone" in multi_frame.columns
    assert before_count == multi_frame.session.count()  # same number of sessions.

    after_shape = multi_frame.loc[Timestamp("2020-12-08 09:36:12+00:00")].shape
    assert before_shape[0] == after_shape[0]  # same number of records
    assert before_shape[1] + 1 == after_shape[1]  # number of columns + 1(hr_zone)



from runpandas.reader import _read_file as read_file  # noqa
from runpandas.reader import _read_dir as read_dir  # noqa
from runpandas.reader import _read_dir_aggregate as read_dir_aggregate  # noqa
from runpandas.io.strava._parser import read_strava  # noqa
from runpandas.io.strava._client import StravaClient  # noqa
from runpandas.io.nikerun._parser import read_nikerun  # noqa
from runpandas.io.nikerun._parser import read_dir_nikerun  # noqa
from runpandas.datasets.utils import activity_examples  # noqa
from runpandas.datasets.schema import FileTypeEnum  # noqa

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

__all__ = [
    "__version__",
    "read_file",
    "read_dir",
    "read_dir_aggregate",
    "read_strava",
    "StravaClient",
    "read_nikerun",
    "read_dir_nikerun",
    "activity_examples",
]



"""
Module contains reading logic for several formats of training sources
"""

from pathlib import Path
import pandas as pd
from runpandas import _utils as utils
from runpandas import exceptions

MODULE_CACHE = {}


def _read_file(filename, to_df=False, **kwargs):
    """
    Parameters
    ----------
        filename : str, The path to a training file.
        to_df : bool, optional
             Return a obj:`runpandas.Activity` if `to_df=True`, otherwise
             a :obj:`pandas.DataFrame` will be returned. Defaults to False.
        **kwargs :
        Keyword args to be passed to the `read` method accordingly to the
        file format.

    Returns
    -------
    Return a obj:`runpandas.Activity` if `to_df=True`, otherwise
             a :obj:`pandas.DataFrame` will be returned.

    """

    if not utils.file_exists(filename):
        raise IOError("%s does not exist" % filename)
    if not utils.is_valid(filename):
        raise exceptions.InvalidFileError(
            "File {filename} with invalid filetype.".format(**locals())
        )
    _, ext = utils.splitext_plus(filename)
    module = _import_module(ext[1:])
    return module.read(filename, to_df, **kwargs)


def _import_module(mod_name):
    """Find custom reading module to execute"""
    mod = MODULE_CACHE.get(mod_name, None)
    if mod is None:
        try:
            MODULE_CACHE[mod_name] = __import__(
                "runpandas.io.%s" % mod_name, fromlist=["runpandas", "io"]
            )
        except ImportError:
            raise ImportError("%s is not a support file type." % mod_name)
        else:
            mod = MODULE_CACHE.get(mod_name)
    return mod


def _read_dir(dirname, to_df=False, **kwargs):
    """

    Parameters
    ----------
        dirname : str, The path to a directory with training files.
        to_df : bool, optional
             Return a obj:`runpandas.Activity` if `to_df=True`, otherwise
             a :obj:`pandas.DataFrame` will be returned. Defaults to False.
        **kwargs : Keyword args to be passed to the `read_dir` method

    Returns
    -------
    Return a list of obj:`runpandas.Activity` if `to_df=True`, otherwise
             a :obj:`pandas.DataFrame` will be returned.

    """
    path_dir = Path(dirname)

    assert path_dir.is_dir()

    for path_file in path_dir.iterdir():
        if path_file.is_dir():
            continue

        yield _read_file(filename=path_file, to_df=to_df, kwargs=kwargs)


def _read_dir_aggregate(dirname, **kwargs):
    """
    Read all supported container files from a supplied directory
    as `runpandas.Activity` dataframes, and aggregate them
    to the same session as a `pandas.MultiIndex` activity dataframe.

    Parameters
    ----------
        dirname : str, The path to a directory with training files.
             Return a obj:`runpandas.Activity` if `to_df=True`, otherwise
             a :obj:`pandas.DataFrame` will be returned. Defaults to False.
        **kwargs : Keyword args to be passed to the `read_dir` method

    Returns
    -------
    Return a  :obj:`runpandas.Activity` split into sessions based
    on the `pandas.MultiIndex` with the date/time of the activity
    as first level and the second the timestamps for each record.
    """
    activities = []
    activity_keys = []
    for activity in _read_dir(dirname=dirname, to_df=False, **kwargs):
        activities.append(activity)
        activity_keys.append(activity.start)
    if len(activities) > 0:
        multi_frame = pd.concat(
            activities, keys=activity_keys, names=["start", "time"], axis=0
        )
        return multi_frame

    return None



"""
Test module for runpandas frame types (i.e. Activity)
"""


import os
import pytest
from pandas import Timedelta, Timestamp
from runpandas import reader

pytestmark = pytest.mark.stable


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


def test_ellapsed_time_frame(dirpath):

    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    frame_gpx = reader._read_file(gpx_file, to_df=False)
    reset_gpx_file = frame_gpx.reset_index()
    with pytest.raises(AttributeError):
        _ = reset_gpx_file.ellapsed_time

    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    frame_gpx = reader._read_file(gpx_file, to_df=False)
    assert frame_gpx.ellapsed_time == Timedelta("0 days 01:25:27")

    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    frame_tcx = reader._read_file(tcx_file, to_df=False)
    assert frame_tcx.ellapsed_time == Timedelta("0 days 00:33:11")

    fit_file = os.path.join(dirpath, "fit", "garmin-fenix-5-basic.fit")
    frame_fit = reader._read_file(fit_file, to_df=False)
    assert frame_fit.ellapsed_time == Timedelta("0 days 00:00:57")


def test_moving_time_frame(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    frame_gpx = reader._read_file(gpx_file, to_df=False)
    reset_gpx_file = frame_gpx.reset_index()
    with pytest.raises(AttributeError):
        _ = reset_gpx_file.moving_time

    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    frame_without_moving = reader._read_file(gpx_file, to_df=False)
    with pytest.raises(AttributeError):
        _ = frame_without_moving.moving_time

    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    frame_gpx = reader._read_file(gpx_file, to_df=False)
    frame_gpx["distpos"] = frame_gpx.compute.distance(correct_distance=False)
    frame_gpx["speed"] = frame_gpx.compute.speed(from_distances=True)
    frame_gpx_only_moving = frame_gpx.only_moving()
    assert frame_gpx_only_moving.moving_time == Timedelta("0 days 01:14:46")

    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    frame_tcx = reader._read_file(tcx_file, to_df=False)
    frame_tcx["distpos"] = frame_tcx.compute.distance(correct_distance=False)
    frame_tcx["speed"] = frame_tcx.compute.speed(from_distances=True)
    frame_tcx = frame_tcx.only_moving()
    assert frame_tcx.moving_time == Timedelta("0 days 00:33:05")

    fit_file = os.path.join(dirpath, "fit", "garmin-fenix-5-basic.fit")
    frame_fit = reader._read_file(fit_file, to_df=False)
    frame_fit_only_moving = frame_fit.only_moving()
    assert frame_fit_only_moving.moving_time == Timedelta("0 days 00:00:55")


def test_distance_frame(dirpath):

    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    frame_no_distance = reader._read_file(gpx_file, to_df=False)
    with pytest.raises(KeyError):
        _ = frame_no_distance.distance

    frame_gpx = reader._read_file(gpx_file, to_df=False)
    frame_gpx["distpos"] = frame_gpx.compute.distance(correct_distance=False)
    assert round(frame_gpx.distance, 2) == 12594.65

    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    frame_tcx = reader._read_file(tcx_file, to_df=False)
    assert round(frame_tcx.distance, 2) == 4686.31

    fit_file = os.path.join(dirpath, "fit", "garmin-fenix-5-basic.fit")
    frame_fit = reader._read_file(fit_file, to_df=False)
    assert round(frame_fit.distance, 2) == 157.56

    tcx_file = os.path.join(dirpath, "tcx", "stopped_example.tcx")
    activity_tcx = reader._read_file(tcx_file, to_df=False)
    assert round(activity_tcx.distance, 2) == 12007.99


def test_mean_speed_frame(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    frame_no_speed = reader._read_file(gpx_file, to_df=False)
    with pytest.raises(AttributeError):
        _ = frame_no_speed.mean_speed()

    frame_gpx = reader._read_file(gpx_file, to_df=False)
    frame_gpx["distpos"] = frame_gpx.compute.distance(correct_distance=False)
    frame_gpx["speed"] = frame_gpx.compute.speed(from_distances=True)
    frame_gpx_only_moving = frame_gpx.only_moving()

    # Calculate the mean speed with only moving  and smoothing using speed (m/s)
    assert (
        frame_gpx_only_moving.mean_speed(only_moving=True, smoothing=True)
    ) == 2.7966895287728653
    # Calculate the mean speed with only moving  and no smoothing (total distance)
    assert (
        frame_gpx_only_moving.mean_speed(only_moving=True, smoothing=False)
    ) == 2.7966895287728653

    # Calculate the mean speed with all data  and no smoothing (total distance)
    assert (
        frame_gpx_only_moving.mean_speed(only_moving=False, smoothing=False)
    ) == 2.4565339871605874
    # Calculate the mean speed with all data  and smoothing (total distance)
    assert (
        frame_gpx_only_moving.mean_speed(only_moving=False, smoothing=True)
    ) == 2.4565339871605874


def test_mean_heartrate_frame(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    frame = reader._read_file(gpx_file, to_df=False)
    frame_no_hr = frame.drop(["hr"], axis=1)
    with pytest.raises(AttributeError):
        _ = frame_no_hr.mean_heart_rate()

    frame_gpx = reader._read_file(gpx_file, to_df=False)
    frame_gpx["distpos"] = frame_gpx.compute.distance(correct_distance=False)
    frame_gpx["speed"] = frame_gpx.compute.speed(from_distances=True)

    frame_gpx_only_moving = frame_gpx.only_moving()

    assert (
        frame_gpx_only_moving.mean_heart_rate(only_moving=False)
    ) == 151.22097819681792
    assert (
        frame_gpx_only_moving.mean_heart_rate(only_moving=True)
    ) == 151.90203327171903


def test_mean_cadence_frame(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    frame = reader._read_file(gpx_file, to_df=False)
    frame_no_cadence = frame.drop(["cad"], axis=1)
    with pytest.raises(AttributeError):
        _ = frame_no_cadence.mean_cadence()

    frame_gpx = reader._read_file(gpx_file, to_df=False)
    frame_gpx["distpos"] = frame_gpx.compute.distance(correct_distance=False)
    frame_gpx["speed"] = frame_gpx.compute.speed(from_distances=True)

    frame_gpx_only_moving = frame_gpx.only_moving()

    assert (frame_gpx_only_moving.mean_cadence(only_moving=False)) == 84.99764289923394
    assert (frame_gpx_only_moving.mean_cadence(only_moving=True)) == 85.96118299445472


def test_mean_pace_frame(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    frame = reader._read_file(gpx_file, to_df=False)
    with pytest.raises(AttributeError):
        _ = frame.mean_pace()

    frame_gpx = reader._read_file(gpx_file, to_df=False)
    frame_gpx["distpos"] = frame_gpx.compute.distance(correct_distance=False)
    frame_gpx["speed"] = frame_gpx.compute.speed(from_distances=True)

    frame_gpx_only_moving = frame_gpx.only_moving()

    # Calculate the mean pace with only moving  and smoothing using speed (m/s)
    assert (
        frame_gpx_only_moving.mean_pace(
            only_moving=True, smoothing=True
        ).total_seconds()
    ) == pytest.approx(
        Timedelta("0 days 00:00:00.357565610").total_seconds(), rel=0.001
    )

    # Calculate the mean pace with only moving  and no smoothing (total distance)
    assert (
        frame_gpx_only_moving.mean_pace(
            only_moving=True, smoothing=False
        ).total_seconds()
    ) == pytest.approx(Timedelta("0 days 00:00:0.357565610").total_seconds(), rel=0.001)

    # Calculate the mean pace with all data  and no smoothing (total distance)
    assert (
        frame_gpx_only_moving.mean_pace(
            only_moving=False, smoothing=False
        ).total_seconds()
    ) == pytest.approx(Timedelta("0 days 00:00:00.407078").total_seconds(), rel=0.001)
    # Calculate the mean pace with all data  and smoothing (total distance)
    assert (
        frame_gpx_only_moving.mean_pace(
            only_moving=False, smoothing=True
        ).total_seconds()
    ) == pytest.approx(Timedelta("0 days 00:00:00.407078").total_seconds(), rel=0.001)


def convert_pace_secmeters2minkms(seconds):
    pace_min = int((seconds * 1000) / 60)
    pace_sec = int(seconds * 1000 - (pace_min * 60))
    total_seconds = (pace_min * 60) + pace_sec
    return Timedelta(seconds=total_seconds)


def test_full_tcx_activity(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    frame_tcx = reader._read_file(tcx_file, to_df=False)

    # test_hr_values_are_correct
    assert frame_tcx["hr"].iloc[-1] == 180
    assert frame_tcx["hr"].iloc[0] == 62

    # test_altitude_points_are_correct
    assert frame_tcx["alt"].iloc[-1] == 166.4453125
    assert frame_tcx["alt"].iloc[0] == 178.942626953

    # test_time_values_are_correct
    assert frame_tcx.index[-1] == Timedelta("0 days 00:33:11")
    assert frame_tcx.index[0] == Timedelta("0 days 00:00:00")

    # test_time_values_are_correct
    assert frame_tcx.start == Timestamp("2012-12-26 21:29:53+00:00")

    # test_latitude_is_correct
    assert frame_tcx["lat"].iloc[0] == pytest.approx(35.951880198, 0.01)
    assert frame_tcx["lon"].iloc[0] == -79.0931872185

    # test_distance_is_correct
    assert frame_tcx.distance == 4686.31103516

    # test_duration_is_correct (we don't consider fraction time)
    assert frame_tcx.ellapsed_time.total_seconds() == 1991

    # test_hr_max
    assert frame_tcx["hr"].max() == 189

    # test_hr_min
    assert frame_tcx["hr"].min() == 60

    # test_hr_avg
    assert int(frame_tcx["hr"].mean()) == 156

    # test_speed
    frame_tcx["distpos"] = frame_tcx.compute.distance()
    frame_tcx["speed"] = frame_tcx.compute.speed(from_distances=True)
    assert frame_tcx.mean_speed() == pytest.approx(2.3545989706033197, 0.01)

    # test_pace (converted to seconds) "07:04"
    pace_min_km = convert_pace_secmeters2minkms(frame_tcx.mean_pace().total_seconds())
    assert pace_min_km == Timedelta("0 days 00:07:04")

    # test_altitude_avg_is_correct
    assert frame_tcx["alt"].mean() == 172.02005618422717

    # test_altitude_max_is_correct
    assert frame_tcx["alt"].max() == 215.95324707

    # test_altitude_min_is_correct
    assert frame_tcx["alt"].min() == 157.793579102

    # test_ascent_is_correct
    assert frame_tcx["alt"].ascent.sum() == pytest.approx(153.80981445000003, 0.01)

    # test_descent_is_correct
    assert frame_tcx["alt"].descent.sum() == pytest.approx(-166.30712890300003, 0.01)

    # test_distance_values_are_correct
    frame_tcx["dist"] = frame_tcx["distpos"].to_distance()
    assert frame_tcx["dist"].fillna(0).iloc[0] == 0.0  # (NaN number for position 0)
    assert frame_tcx["dist"].iloc[-1] == pytest.approx(
        4688.006550471207, 0.01
    )  # 4686.31103516 (Precision?)


def test_full_gpx_activity(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "run.gpx")
    frame_gpx = reader._read_file(gpx_file, to_df=False)

    # testActivityStartTime
    assert frame_gpx.start == Timestamp("2017-05-27 08:13:01+00:00")

    # testActivityTotalDistance
    frame_gpx["distpos"] = frame_gpx.compute.distance()
    frame_gpx["dist"] = frame_gpx["distpos"].to_distance()
    assert frame_gpx["dist"].fillna(0).iloc[0] == 0.0  # (NaN number for position 0)
    assert frame_gpx["dist"].iloc[-1] == 4806.843188885856

    # test_duration_is_correct (we don't consider fraction time)
    assert frame_gpx.ellapsed_time.total_seconds() == 1423

    # test_speed
    frame_gpx["speed"] = frame_gpx.compute.speed(from_distances=True)
    assert frame_gpx.mean_speed() == 3.3779642929626545

    # test_pace (converted to seconds) "00:04:55"
    pace_min_km = convert_pace_secmeters2minkms(frame_gpx.mean_pace().total_seconds())
    assert pace_min_km == Timedelta("0 days 00:04:56")

    # test_max_speed
    assert frame_gpx["speed"].max() == pytest.approx(5.458744217718473, 0.01)

    # test_ascent_is_correct
    assert frame_gpx["alt"].ascent.sum() == 50.90000000000001

    # test_descent_is_correct
    assert frame_gpx["alt"].descent.sum() == -50.20000000000001


def test_full_fit_activity(dirpath):
    fit_file = os.path.join(dirpath, "fit", "run.fit")
    frame_fit = reader._read_file(fit_file, to_df=False)

    # testActivityStartTime
    assert frame_fit.start == Timestamp("2019-09-14 15:22:05+0000")

    # test_distance_is_correct
    assert frame_fit.distance == 5839.77

    # test_duration_is_correct (we don't consider fraction time)
    assert frame_fit.ellapsed_time.total_seconds() == 3167.0

    # test_hr_avg
    assert int(frame_fit["hr"].mean()) == 129

    # test_speed
    assert frame_fit.mean_speed() == 1.8114262709188504

    # test average cadence
    assert int(frame_fit["cad"].mean()) == 63

    # test average temperature
    assert int(frame_fit["temp"].mean()) == 26


def test_full_tcx_2_activity(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "run.tcx")
    frame_tcx = reader._read_file(tcx_file, to_df=False)

    # testActivityStartTime
    assert frame_tcx.start == Timestamp("2017-05-27 08:13:01+00:00")

    # testActivityTotalDistance
    frame_tcx["distpos"] = frame_tcx.compute.distance()
    frame_tcx["dist"] = frame_tcx["distpos"].to_distance()
    assert frame_tcx["dist"].fillna(0).iloc[0] == 0.0  # (NaN number for position 0)
    assert frame_tcx["dist"].iloc[-1] == pytest.approx(4806.843188885856, 0.01)

    # test_duration_is_correct (we don't consider fraction time)
    assert frame_tcx.ellapsed_time.total_seconds() == 1423

    # test_speed
    frame_tcx["speed"] = frame_tcx.compute.speed(from_distances=True)
    assert frame_tcx.mean_speed() == pytest.approx(3.3779642929626545, 0.01)

    # test_pace (converted to seconds) "00:04:55"
    pace_min_km = convert_pace_secmeters2minkms(frame_tcx.mean_pace().total_seconds())
    assert pace_min_km == Timedelta("0 days 00:04:56")

    # test_max_speed
    assert frame_tcx["speed"].max() == pytest.approx(5.458744217718473, 0.01)

    # test_ascent_is_correct
    assert frame_tcx["alt"].ascent.sum() == 50.90000000000001

    # test_descent_is_correct
    assert frame_tcx["alt"].descent.sum() == -50.20000000000001


def test_full_tcx_garmin_activity(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "run_garmin.tcx")
    frame_tcx = reader._read_file(tcx_file, to_df=False)

    # testActivityStartTime
    assert frame_tcx.start == Timestamp("2018-11-25 07:20:49+00:00")

    # test_hr_values_are_correct
    assert frame_tcx["hr"].iloc[-1] == 143
    assert frame_tcx["hr"].iloc[0] == 112

    # test_altitude_points_are_correct
    assert frame_tcx["alt"].iloc[-1] == 5.199999809265137
    assert frame_tcx["alt"].iloc[0] == 4.800000190734863

    # test_time_values_are_correct
    assert frame_tcx.index[-1] == Timedelta("0 days 00:13:43")
    assert frame_tcx.index[0] == Timedelta("0 days 00:00:00")

    # test_latitude_is_correct
    assert frame_tcx["lat"].iloc[0] == 46.15864304825664
    assert frame_tcx["lon"].iloc[0] == -1.1469579208642244

    # test_distance_is_correct
    assert frame_tcx.distance == 2153.669921875

    # test_duration_is_correct (we don't consider fraction time)
    assert frame_tcx.ellapsed_time.total_seconds() == 823

    # test_hr_max
    assert frame_tcx["hr"].max() == 153

    # test_hr_min
    assert frame_tcx["hr"].min() == 112

    # test_hr_avg
    assert int(frame_tcx["hr"].mean()) == 142

    # test_speed USING GARMIN ESTIMATED SPEED
    assert frame_tcx.mean_speed() == 2.6589513960614672

    # test_pace (converted to seconds) "06:16"
    pace_min_km = convert_pace_secmeters2minkms(frame_tcx.mean_pace().total_seconds())
    assert pace_min_km == Timedelta("0 days 00:06:16")

    # test_altitude_avg_is_correct
    assert frame_tcx["alt"].mean() == 5.210884340766336

    # test_altitude_max_is_correct
    assert frame_tcx["alt"].max() == pytest.approx(6.599999904632568, 0.01)

    # test_altitude_min_is_correct
    assert frame_tcx["alt"].min() == 4.400000095367432

    # test_ascent_is_correct
    assert frame_tcx["alt"].ascent.sum() == pytest.approx(20.59999990463257, 0.01)

    # test_descent_is_correct
    assert frame_tcx["alt"].descent.sum() == pytest.approx(-20.200000286102295, 0.01)

    # test_distance_values_are_correct
    assert frame_tcx["dist"].fillna(0).iloc[0] == 2.5299999713897705
    assert frame_tcx["dist"].iloc[-1] == 2153.669921875



"""
Tools for parsing Nike Run JSON files.
"""
import os
import datetime
import json
from pathlib import Path
import numpy as np
import pandas as pd
from pandas import TimedeltaIndex
from runpandas import _utils as utils
from runpandas import exceptions
from runpandas.types import Activity, columns

# According to Garmin, all times are stored in UTC.
DATETIME_FMT = "%Y-%m-%dT%H:%M:%SZ"

COLUMNS_SCHEMA = {
    "elevation": columns.Altitude,
    "heart_rate": columns.HeartRate,
    "longitude": columns.Longitude,
    "latitude": columns.Latitude,
}


def __is_nikerun_valid(file_path):
    """Check if it is a valid format for activity files.

    Parameters
    ----------
    file_path : str
        Path to the file to be read.

    Returns
    -------
    It returns True if the it is a valid format for activities handling.
    """
    _, ext = os.path.splitext(file_path)
    if ext not in [".json"]:
        return False

    with open(file_path) as json_file:
        activity = json.load(json_file)
        if not activity.get("metrics"):
            return False

        metrics = [
            metric["type"]
            for metric in activity["metrics"]
            if metric["type"] in ["latitude", "longitude"]
        ]
        if len(metrics) != 2:
            return False

    return True


def __update_single_metrics(epochs, data):
    counter = 0
    data_values = [np.nan] * len(epochs)
    for idx, epoch in enumerate(epochs):
        while epoch >= data[counter]["end_epoch_ms"]:
            if counter == len(data) - 1:
                break
            data_values[idx] = data[counter]["value"]
            counter += 1

    return data_values


def __update_acummulative_metrics(epochs, data):
    counter = 0
    data_values = [np.nan] * len(epochs)
    for idx, epoch in enumerate(epochs):
        data_cumm = []
        while epoch >= data[counter]["end_epoch_ms"]:
            if counter == len(data) - 1:
                break
            data_cumm.append(data[counter]["value"])
            counter += 1
        if len(data_cumm) > 0:
            data_values[idx] = sum(data_cumm)

    return data_values


def __nikerun_streams(metrics):
    streams = dict()
    stream_types = dict()
    final_streams = dict()
    for metric in metrics:
        type_metric = metric["type"]
        unit = metric["unit"]
        streams[type_metric] = metric["values"]
        stream_types[type_metric] = unit

    # get latitude and longitude, build them and corresponding timestamp values
    latitude_data, longitude_data = streams["latitude"], streams["longitude"]
    latitude_values, longitude_values, epoch_ms, epoch_datetime = [], [], [], []
    for lat, lon in zip(latitude_data, longitude_data):
        if lat["start_epoch_ms"] != lon["start_epoch_ms"]:
            raise ValueError("\tThe latitude and longitude data is out of order")
        latitude_values.append(lat["value"])
        longitude_values.append(lon["value"])
        epoch_ms.append(lat["start_epoch_ms"])
        epoch_datetime.append(
            datetime.datetime.utcfromtimestamp(lat["start_epoch_ms"] / 1000)
        )

    final_streams["latitude"] = latitude_values
    final_streams["longitude"] = longitude_values
    final_streams["time"] = epoch_datetime

    final_streams["elevation"] = __update_single_metrics(epoch_ms, streams["elevation"])
    if "heart_rate" in streams:
        final_streams["heart_rate"] = __update_single_metrics(
            epoch_ms, streams["heart_rate"]
        )

    if "calories" in streams:
        final_streams["calories"] = __update_acummulative_metrics(
            epoch_ms, streams["calories"]
        )
    if "steps" in streams:
        final_streams["steps"] = __update_acummulative_metrics(
            epoch_ms, streams["steps"]
        )
    if "nikefuel" in streams:
        final_streams["nikefuel"] = __update_acummulative_metrics(
            epoch_ms, streams["nikefuel"]
        )

    return final_streams


def gen_records(file_path):
    with open(file_path) as json_file:
        activity = json.load(json_file)
        streams = __nikerun_streams(activity.get("metrics"))
        return streams


def read_nikerun(file_path, to_df=False, **kwargs):
    """
    This method loads a NikeRun API response in JSON file into a Pandas DataFrame
    or runpandas Activity.
    Column names are translated to runpandas terminology
    (e.g. "HeartRate" > "heart_rate").
    Datetimes indexes are replaced by time offsets.
    All NaN rows are removed.

    Parameters
    ----------
        file_path : str, The path to a training file.
        to_df : bool, optional
             Return a obj:`runpandas.Activity` if `to_df=True`,
              otherwise a :obj:`pandas.DataFrame` will be returned.
              Defaults to False.
        **kwargs :
        Keyword args to be passed to the `read` method
              accordingly to the file format.
    Returns
    -------
    Return a obj:`runpandas.Activity` if `to_df=True`, otherwise
             a :obj:`pandas.DataFrame` will be returned.
    """
    if not utils.file_exists(file_path):
        raise IOError("%s does not exist" % file_path)
    if not __is_nikerun_valid(file_path):
        raise exceptions.InvalidFileError(
            "File {file_path} with invalid filetype.".format(**locals())
        )

    data = pd.DataFrame.from_records(gen_records(file_path))
    times = data.pop("time")  # should always be there
    data = data.astype("float64", copy=False)  # try and make numeric
    data.columns = map(utils.camelcase_to_snakecase, data.columns)
    timestamps = pd.to_datetime(times, format=DATETIME_FMT, utc=True)

    timeoffsets = timestamps - timestamps[0]
    timestamp_index = TimedeltaIndex(timeoffsets, unit="s", name="time")
    data.index = timestamp_index
    data.dropna(axis=1, how="all", inplace=True)

    if to_df:
        return data

    return Activity(data, cspecs=COLUMNS_SCHEMA, start=timestamps[0])


def read_dir_nikerun(dirname, **kwargs):
    """
    Read all NikeRun JSON container files from a supplied directory
    as `runpandas.Activity` dataframes, and aggregate them
    to the same session as a `pandas.MultiIndex` activity dataframe.

    Parameters
    ----------
        dirname : str, The path to a directory with training files.
             Return a obj:`runpandas.Activity` if `to_df=True`, otherwise
             a :obj:`pandas.DataFrame` will be returned. Defaults to False.
        **kwargs : Keyword args to be passed to the `read_dir` method

    Returns
    -------
    Return a  :obj:`runpandas.Activity` split into sessions based
    on the `pandas.MultiIndex` with the date/time of the activity
    as first level and the second the timestamps for each record.
    """
    activities = []
    activity_keys = []

    path_dir = Path(dirname)

    assert path_dir.is_dir()

    for path_file in path_dir.iterdir():
        if path_file.is_dir():
            continue

        activity = read_nikerun(path_file, to_df=False, **kwargs)
        activities.append(activity)
        activity_keys.append(activity.start)
    if len(activities) > 0:
        multi_frame = pd.concat(
            activities, keys=activity_keys, names=["start", "time"], axis=0
        )
        return multi_frame

    return None



"""
This is a module for extending pandas dataframes with the runpandas toolbox
"""

import pandas as pd
import numpy as np
from haversine import haversine as hv
from haversine import Unit
from runpandas import exceptions
from runpandas._utils import special_column
from runpandas.types import Activity
from runpandas.types import columns


@pd.api.extensions.register_dataframe_accessor("compute")
class MetricsAcessor(object):
    """
    Dataframe Accessor to compute all available metrics from an activity

    Raises
    ------
    AttributeError
        if at least one of the needed columns is missing.

    """

    def __init__(self, activity):
        self._validate(activity)
        self._activity = activity

    @staticmethod
    def _validate(activity):
        assert isinstance(activity, Activity), "Activity is needed for this accessor"

        if not isinstance(activity.index, (pd.DatetimeIndex, pd.TimedeltaIndex)):
            raise AttributeError(
                "An activity needs a DataFrame Index in DateTimeIndex format"
            )

    def __correct_distance(self, distance_series, col_alt="alt"):
        if col_alt not in self._activity.columns:
            raise exceptions.RequiredColumnError(col_alt)

        self._activity["shift_alt"] = self._activity[col_alt].shift(1)
        self._activity["alt_dif"] = self._activity.apply(
            lambda x: x["shift_alt"] - x[col_alt], axis=1
        )
        self._activity["distance_haversine"] = distance_series
        distance_corrected = self._activity.apply(
            lambda x: np.sqrt(x["distance_haversine"] ** 2 + (x["alt_dif"]) ** 2),
            axis=1,
        )
        self._activity.drop(
            ["shift_alt", "alt_dif", "distance_haversine"], axis=1, inplace=True
        )

        return distance_corrected

    @special_column(required_columns=("lat", "lon"), name="distpos")
    def distance(self, correct_distance=False, to_special_column=True, **kwargs):
        """
        Calculates the distance in meters using haversine distance formula on an Activity frame.

        Parameters
        ----------
        correct_distance: bool, optional
            It computes the distance corrected by the altitude. default is False.

        to_special_column: bool, optional
            It converts the distance calculated (`pandas.Series`) to special runpandas
            distance cummulative column (`runpandas.types.columns.DistancePerPosition`).
            Default is True.

        **kwargs: Keyword args to be passed to the `haversine` method

        Returns
        -------
        haversine_dist: pandas.Series or runpandas.types.columns.DistancePerPosition
            A Series of floats representing the distance in meters
            with the same index of the accessed activity object.

        """
        self._activity["point"] = self._activity.apply(
            lambda x: (x["lat"], x["lon"]), axis=1
        )
        self._activity["point_next"] = self._activity["point"].shift(1)
        self._activity.loc[self._activity["point_next"].isna(), "point_next"] = None
        haversine_dist = self._activity.apply(
            lambda x: hv(x["point"], x["point_next"], unit=Unit.METERS)
            if x["point_next"] is not None
            else float("nan"),
            axis=1,
        )
        self._activity.drop(["point_next", "point"], axis=1, inplace=True)

        if correct_distance:
            haversine_dist = self.__correct_distance(haversine_dist)

        if to_special_column:
            haversine_dist = columns.DistancePerPosition(haversine_dist)

        return haversine_dist

    @special_column(required_columns=(), name="speed")
    def speed(self, from_distances=False, to_special_column=True, **kwargs):
        """
        Calculates the speed in meters using an Activity frame.

        Parameters
        ----------
        from_distances: bool, optional
            Should the speeds be calculated from the distance recordings
            instead of taken from the speed recordings directly? Default is False.

        to_special_column: convert the distance calculated (`pandas.Series`)
            to special runpandas speed column (`runpandas.types.columns.Speed`).
            Default is True.

        **kwargs: Keyword args to be passed to the speed building method

        Returns
        -------
        speed: `pandas.Series` or `runpandas.types.columns.Speed`
            A Series of floats representing the speed in meters
            with the same index of the accessed activity object.

        """
        if from_distances:
            if "distpos" not in self._activity.columns:
                raise exceptions.RequiredColumnError("distpos")

            time_diff = (
                self._activity.index.to_series().diff().fillna(self._activity.index[0])
            ) / np.timedelta64(1, "s")
            speed = self._activity["distpos"] / time_diff
        else:
            if "speed" not in self._activity.columns:
                raise exceptions.RequiredColumnError("speed")
            speed = self._activity["speed"]

        if to_special_column:
            speed = columns.Speed(speed)

        return speed

    @special_column(required_columns=("alt",), name="vam")
    def vertical_speed(self, to_special_column=True, **kwargs):
        """
        Calculates the vertical climbing speed (VAM) in meters using an Activity frame.

        Parameters
        ----------
        to_special_column: convert the distance calculated (`pandas.Series`)
            to special runpandas VAM column (`runpandas.types.columns.VAM`).
            Default is True.

        **kwargs: Keyword args to be passed to the VAM build method

        Returns
        -------
        vam: `pandas.Series` or `runpandas.types.columns.VAM`
            A Series of floats representing the vertical altitude speed in meters
            with the same index of the accessed activity object.

        """
        time_diff = (
            self._activity.index.to_series().diff().fillna(self._activity.index[0])
        ) / np.timedelta64(1, "s")
        dvert = self._activity["alt"].diff()

        vam = dvert / time_diff
        if to_special_column:
            vam = columns.VAM(vam)

        return vam

    @special_column(required_columns=("alt", "dist"), name="grad")
    def gradient(self, to_special_column=True, **kwargs):
        """
        Calculates the gradient ratio from an Activity frame.

        Parameters
        ----------
        to_special_column: convert the distance calculated (`pandas.Series`)
            to special runpandas Gradient column (`runpandas.types.columns.Gradient`).
            Default is True.

        **kwargs: Keyword args to be passed to the Gradient build method

        Returns
        -------
        grad: `pandas.Series` or `runpandas.types.columns.Gradient`
            A Series of floats representing the vertical altitude speed in meters
            with the same index of the accessed activity object.

        """
        alt = self._activity["alt"].diff()
        dist = self._activity["dist"].diff()

        grad = alt / dist
        if to_special_column:
            grad = columns.Gradient(rise=alt, run=dist)

        return grad

    @special_column(required_columns=("speed",), name="pace")
    def pace(self, to_special_column=True, **kwargs):
        """
        Calculates the pace (the time that it takes to cover distances in your activities).

        Parameters
        ----------
        to_special_column: convert the pace calculated (`pandas.Series`)
            to special runpandas Pace column (`runpandas.types.columns.Pace`).
            Default is True.

        **kwargs: Keyword args to be passed to the Pace build method

        Returns
        -------
        pace: `pandas.Series` or `runpandas.types.columns.Pace`
            A Series of floats representing the pace in meters per second
            with the same index of the accessed activity object.

        """
        speed = self._activity["speed"]

        pace = pd.to_timedelta(1 / speed, unit="s")
        # pace = (1 / speed).apply(pd.Timedelta, args=('s',))

        if to_special_column:
            pace = columns.Pace(pace)

        return pace

    @special_column(required_columns=("hr",))
    def heart_zone(self, bins, labels, **kwargs):
        """
        Returns a `pandas.Series` with the training zone labels for each heart rate record.
        This method uses the pandas.cut() method.
        Nan will be returned for values that are not in any of the bins.

        Parameters
        ----------

        bins: Left and right bounds for each zone.
                An example of valid bins are [0, 100, 140, 160, 999].

        labels: Specifies the labels for the zones.
                Must be the same length as the resulting zones.
                Example of valid labels is ["Z1", "Z2", "Z3", "Z4", "Z5"].

        **kwargs: Keyword args to be passed to the heart_zone build method

        Returns
        -------
        series:  `pandas.Series` with the zone label for each value.
        """
        bins_series = pd.cut(self._activity["hr"], bins=bins, labels=labels)
        return pd.Series(bins_series, index=bins_series.index, name="hr_zone")

    @special_column(required_columns=("hr",), name="time_in_zone")
    def time_in_zone(self, bins, labels, **kwargs):
        """
        Returns a `pandas.Series` with the values counts in timedelta for each
        heart training zone.
        This method uses the pandas.Series.values_count() method.

        Parameters
        ----------

        bins: Left and right bounds for each zone.
                An example of valid bins are [0, 100, 140, 160, 999].

        labels: Specifies the labels for the zones.
                Must be the same length as the resulting zones.
                Example of valid labels is ["Z1", "Z2", "Z3", "Z4", "Z5"].

        **kwargs: Keyword args to be passed to the time_in_zone build method

        Returns
        -------
        series: A `pandas.Series` with the values count in `pandas.Timedelta` for each zones.
        """
        hr_zones = self.heart_zone(bins, labels).to_frame()
        hr_zones["time_diff"] = (
            hr_zones.index.to_series().diff().fillna(hr_zones.index[0])
        ) / np.timedelta64(1, "s")
        return pd.to_timedelta(hr_zones.groupby(["hr_zone"]).time_diff.sum(), unit="s")



"""
Test module for FIT reader base module
"""

import os
import pytest
from pandas import DataFrame, TimedeltaIndex, Timedelta, Timestamp
from runpandas import reader
from runpandas import types
import runpandas.io.fit._parser as fit_parser
from fitparse.utils import FitParseError

pytestmark = pytest.mark.stable


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


def test_read_file_fit_no_fit(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "garmin_connect.gpx")
    with pytest.raises(FitParseError):
        fit_parser.read(gpx_file)


@pytest.fixture
def pandas_activity(dirpath):
    fit_file = os.path.join(dirpath, "fit", "garmin-fenix-5-basic.fit")
    return reader._read_file(fit_file, to_df=True)


@pytest.fixture
def runpandas_activity(dirpath):
    fit_file = os.path.join(dirpath, "fit", "garmin-fenix-5-basic.fit")
    return reader._read_file(fit_file, to_df=False)


test_data = [
    (pytest.lazy_fixture("pandas_activity"), "heart_rate", 0, 61),
    (pytest.lazy_fixture("pandas_activity"), "heart_rate", -1, 112),
    (pytest.lazy_fixture("pandas_activity"), "altitude", 0, 2.1999999999999886),
    (pytest.lazy_fixture("pandas_activity"), "altitude", -1, 4.199999999999989),
    (pytest.lazy_fixture("pandas_activity"), "position_lat", 0, 456099128),
    (pytest.lazy_fixture("pandas_activity"), "position_long", 0, -1463077077),
    (pytest.lazy_fixture("pandas_activity"), "distance", -1, 157.56),
    (pytest.lazy_fixture("pandas_activity"), "distance", 0, 0.0),
    (pytest.lazy_fixture("pandas_activity"), "temperature", 0, 25),
    (pytest.lazy_fixture("pandas_activity"), "temperature", -1, 24),
    (pytest.lazy_fixture("pandas_activity"), "speed", 0, 0.0),
    (pytest.lazy_fixture("pandas_activity"), "speed", -1, 2.865),
    (pytest.lazy_fixture("pandas_activity"), "cadence", 0, 0.0),
    (pytest.lazy_fixture("pandas_activity"), "cadence", -1, 88),
    (pytest.lazy_fixture("runpandas_activity"), "hr", 0, 61),
    (pytest.lazy_fixture("runpandas_activity"), "hr", -1, 112),
    (pytest.lazy_fixture("runpandas_activity"), "alt", 0, 2.1999999999999886),
    (pytest.lazy_fixture("runpandas_activity"), "alt", -1, 4.199999999999989),
    (pytest.lazy_fixture("runpandas_activity"), "lat", 0, 38.22978727519512),
    (pytest.lazy_fixture("runpandas_activity"), "lon", 0, -122.63370391912758),
    (pytest.lazy_fixture("runpandas_activity"), "dist", -1, 157.56),
    (pytest.lazy_fixture("runpandas_activity"), "dist", 0, 0.0),
    (pytest.lazy_fixture("runpandas_activity"), "cad", -1, 88),
    (pytest.lazy_fixture("runpandas_activity"), "cad", 0, 0),
    (pytest.lazy_fixture("runpandas_activity"), "temp", -1, 24),
    (pytest.lazy_fixture("runpandas_activity"), "temp", 0, 25),
    (pytest.lazy_fixture("runpandas_activity"), "speed", -1, 2.865),
    (pytest.lazy_fixture("runpandas_activity"), "speed", 0, 0.0),
]


@pytest.mark.parametrize("activity,column,index,expected", test_data)
def test_fit_values(activity, column, index, expected):
    assert activity[column].iloc[index] == expected
    assert activity.index[-1] == Timedelta("0 days 00:00:57")

    if isinstance(activity, types.Activity):
        assert activity.start == Timestamp("2017-06-11 14:34:09+00:00")


def test_read_file_fit_basic_dataframe(dirpath):
    fit_file = os.path.join(dirpath, "fit", "garmin-fenix-5-basic.fit")
    activity = reader._read_file(fit_file, to_df=True)
    assert isinstance(activity, DataFrame)
    assert isinstance(activity.index, TimedeltaIndex)
    assert activity.size == 462
    included_data = set(
        [
            "position_lat",
            "position_long",
            "distance",
            "temperature",
            "altitude",
            "speed",
            "heart_rate",
            "cadence",
            "speed",
            "lap",
            "session",
        ]
    )
    assert included_data <= set(activity.columns.to_list())

    assert "lap" in activity.columns
    assert activity["lap"].max() == 0  # no laps

    assert "session" in activity.columns
    assert activity["session"].max() == 0  # no sessions


def test_read_file_fit_basic_activity(dirpath):
    gpx_file = os.path.join(dirpath, "fit", "garmin-fenix-5-basic.fit")
    activity = reader._read_file(gpx_file, to_df=False)
    assert isinstance(activity, types.Activity)
    assert isinstance(activity.index, TimedeltaIndex)
    assert activity.size == 462

    included_data = set(["lat", "lon", "alt", "cad", "hr", "speed", "temp"])
    assert included_data <= set(activity.columns.to_list())

    assert "lap" in activity.columns
    assert activity["lap"].max() == 0  # no laps

    assert "session" in activity.columns
    assert activity["session"].max() == 0  # no sessions



"""
Tools for pulling and parsing stream data from Strava.
"""
from datetime import timedelta
import pandas as pd
from pandas import TimedeltaIndex
from runpandas import _utils as utils
from runpandas.types import Activity
from runpandas.types import columns
from runpandas.io.strava._client import StravaClient


COLUMNS_SCHEMA = {
    "altitude": columns.Altitude,
    "cadence": columns.Cadence,
    "distance": columns.Distance,
    "heartrate": columns.HeartRate,
    "longitude": columns.Longitude,
    "latitude": columns.Latitude,
    "watts": columns.Power,
}

STREAM_TYPES = [
    "time",
    "latlng",
    "distance",
    "altitude",
    "velocity_smooth",
    "heartrate",
    "cadence",
    "watts",
    "temp",
    "moving",
    "grade_smooth",
]


def gen_records(streams):
    raw_data = dict()
    for key, value in streams.items():
        if key == "latlng":
            latitude, longitude = list(zip(*value.data))
            raw_data["latitude"] = latitude
            raw_data["longitude"] = longitude
        else:
            raw_data[key] = value.data
    return raw_data


def read_strava(
    activity_id,
    client=None,
    to_df=False,
    **kwargs,
):
    """
    This method loads the activity data from Strava into a Pandas DataFrame or
    runpandas Activity.
    Column names are translated to runpandas terminology
    (e.g. "heartrate" > "heart_rate").
    Datetimes indexes are replaced by time offsets.
    All NaN rows are removed.

    Attention: Two API requests are made to the Strava webservice: 1 to
               retrieve activity metadata, 1 to retrieve the raw data ("streams").

    Parameters
    ----------
        activity_id : str, The id of the activity
        access_token: str, The Strava access token
        refresh_token: str, The Strava refresh token, optional
        client_id: int, The Strava client id used for token refresh, optional
        client_secret: str, The strava client secret used for token refresh, optional
        to_df : bool, optional
             Return a obj:`runpandas.Activity` if `to_df=True`, otherwise
             a :obj:`pandas.DataFrame` will be returned. Defaults to False.
        **kwargs :
        Keyword args to be passed to the `read_strava`
    Returns
    -------
    Return a obj:`runpandas.Activity` if `to_df=True`, otherwise
             a :obj:`pandas.DataFrame` will be returned.
    """
    if client is None:
        client = StravaClient()
    client.refresh()

    activity = client.get_activity(activity_id)

    start_datetime = activity.start_date_local
    streams = client.get_activity_streams(
        activity_id=activity_id, types=STREAM_TYPES, series_type="time"
    )

    data = pd.DataFrame(gen_records(streams))

    times = data.pop("time")
    data.columns = map(utils.camelcase_to_snakecase, data.columns)

    def time_to_datetime(time):
        return start_datetime + timedelta(seconds=time)

    timestamps = times.apply(time_to_datetime)
    timeoffsets = timestamps - timestamps[0]
    timestamp_index = TimedeltaIndex(timeoffsets, unit="s", name="time")
    data.index = timestamp_index
    data.dropna(axis=1, how="all", inplace=True)

    if to_df:
        return data

    return Activity(data, cspecs=COLUMNS_SCHEMA, start=timestamps[0])



"""
Tools for parsing Garmin TCX files.
"""
import pandas as pd
from pandas import TimedeltaIndex
from runpandas import _utils as utils
from runpandas import exceptions
from runpandas.types import Activity
from runpandas.types import columns

COLUMNS_SCHEMA = {
    "altitude_meters": columns.Altitude,
    "cadence": columns.Cadence,
    "distance_meters": columns.Distance,
    "heart_rate_bpm": columns.HeartRate,
    "longitude_degrees": columns.Longitude,
    "latitude_degrees": columns.Latitude,
    "speed": columns.Speed,
    "watts": columns.Power,
}

# According to Garmin, all times are stored in UTC.
DATETIME_FMT = "%Y-%m-%dT%H:%M:%SZ"
# Despite what the schema says, there are files out
# in the wild with fractional seconds...
DATETIME_FMT_WITH_FRAC = "%Y-%m-%dT%H:%M:%S.%fZ"


def gen_records(file_path):
    nodes = utils.get_nodes(file_path, ("Trackpoint",), with_root=True)
    root = next(nodes)
    if utils.sans_ns(root.tag) != "TrainingCenterDatabase":
        raise exceptions.InvalidFileError("tcx")

    trackpoints = nodes
    for trkpt in trackpoints:
        yield utils.recursive_text_extract(trkpt)


def read(file_path, to_df=False, **kwargs):
    """
    This method loads a TCX file into a Pandas DataFrame or runpandas Activity.
    Column names are translated to runpandas terminology
    (e.g. "HeartRate" > "heart_rate").
    Datetimes indexes are replaced by time offsets.
    All NaN rows are removed.

    Parameters
    ----------
        filename : str, The path to a training file.
        to_df : bool, optional
             Return a obj:`runpandas.Activity` if `to_df=True`, otherwise
             a :obj:`pandas.DataFrame` will be returned. Defaults to False.
        **kwargs :
        Keyword args to be passed to the `read` method accordingly to the
        file format.
    Returns
    -------
    Return a obj:`runpandas.Activity` if `to_df=True`, otherwise
             a :obj:`pandas.DataFrame` will be returned.
    """
    data = pd.DataFrame.from_records(gen_records(file_path))
    times = data.pop("Time")  # should always be there
    data = data.apply(pd.to_numeric, errors="ignore")  # try and make numeric

    data.columns = map(utils.camelcase_to_snakecase, data.columns)

    try:
        timestamps = pd.to_datetime(times, format=DATETIME_FMT, utc=True)
    except ValueError:
        # bad format, try with fractional seconds
        timestamps = pd.to_datetime(times, format=DATETIME_FMT_WITH_FRAC, utc=True)

    timeoffsets = timestamps - timestamps[0]
    timestamp_index = TimedeltaIndex(timeoffsets, unit="s", name="time")
    data.index = timestamp_index
    data.dropna(axis=1, how="all", inplace=True)

    if to_df:
        return data

    return Activity(data, cspecs=COLUMNS_SCHEMA, start=timestamps[0])



"""
    Helper module for evaluation and display of the summary of training sessions.

 """

import numpy as np
import pandas as pd
from runpandas._utils import convert_pace_secmeters2minkms


def _build_summary_statistics(obj):
    """
    Generate session statistics from a given DataFrame.

    Parameters
    ----------
    obj:  The DataFrame to generate basic commute statistics from.

    Returns:
    --------
    A Dictionary containing the following statistics:
    - Total moving time
    - Average speed
    - Max speed
    - Average moving speed
    - Average cadence running
    - Average cadence running moving
    - Max cadence
    - Average heart rate
    - Average heart rate moving
    - Max heart rate
    - Average pace (per 1 km)
    - Average pace moving (per 1 km)
    - Max pace
    - Average temperature
    - Max temperature
    - Min temperature
    - Total distance
    - Total ellapsed time
    """

    start = obj.start

    try:
        moving_time = obj.moving_time
    except AttributeError:
        moving_time = np.nan

    try:
        mean_speed = obj.mean_speed()
        max_speed = obj["speed"].max()
        mean_pace = convert_pace_secmeters2minkms(obj.mean_pace().total_seconds())
        max_pace = convert_pace_secmeters2minkms(
            obj["speed"].to_pace().min().total_seconds()
        )
    except AttributeError:
        mean_speed = np.nan
        max_speed = np.nan
        mean_pace = np.nan

    try:
        mean_moving_speed = obj.mean_speed(only_moving=True)
        mean_moving_pace = convert_pace_secmeters2minkms(
            obj.mean_pace(only_moving=True).total_seconds()
        )
    except (AttributeError, KeyError):
        mean_moving_speed = np.nan
        mean_moving_pace = np.nan

    try:
        mean_cadence = obj.mean_cadence()
        max_cadence = obj["cad"].max()
    except AttributeError:
        mean_cadence = np.nan
        max_cadence = np.nan

    try:
        mean_moving_cadence = obj.mean_cadence(only_moving=True)
    except (AttributeError, KeyError):
        mean_moving_cadence = np.nan

    try:
        mean_heart_rate = obj.mean_heart_rate()
        max_heart_rate = obj["hr"].max()
    except AttributeError:
        mean_heart_rate = np.nan
        max_heart_rate = np.nan

    try:
        mean_moving_heart_rate = obj.mean_heart_rate(only_moving=True)
    except (AttributeError, KeyError):
        mean_moving_heart_rate = np.nan

    try:
        mean_temperature = obj["temp"].mean()
        min_temperature = obj["temp"].min()
        max_temperature = obj["temp"].max()
    except KeyError:
        mean_temperature = np.nan
        min_temperature = np.nan
        max_temperature = np.nan

    total_distance = obj.distance

    ellapsed_time = obj.ellapsed_time

    row = {k: v for k, v in locals().items() if not k.startswith("__") and k != "obj"}

    return row


def _build_session_statistics(obj):
    """
    Generate session statistics from a given DataFrame.

    Parameters
    ----------
    obj:  The DataFrame to generate basic commute statistics from.

    Returns:
    --------
    A ``pandas.Dataframe`` containing the following statistics:
    - Total moving time
    - Average speed
    - Max speed
    - Average moving speed
    - Average cadence running
    - Average cadence running moving
    - Max cadence
    - Average heart rate
    - Average heart rate moving
    - Max heart rate
    - Average pace (per 1 km)
    - Average pace moving (per 1 km)
    - Max pace
    - Average temperature
    - Max temperature
    - Min temperature
    - Total distance
    - Total ellapsed time
    """
    stats = {key: [value] for key, value in _build_summary_statistics(obj).items()}
    return pd.DataFrame(stats).set_index("start")


def _build_activity_statistics(obj):
    """
    Generate basic statistics from a given pandas Series.

    Parameters
    ----------
    obj:  The DataFrame to generate basic commute statistics from.

    Returns:
    --------

    A Series containing the following statistics:
    - Session times
    - Total distance
    - Total ellapsed time
    - Total moving time
    - Total and average elevation gain
    - Average speed
    - Average moving speed
    - Average pace (per 1 km)
    - Average pace moving (per 1 km)
    - Average cadence running
    - Average cadence running moving
    - Average heart rate
    - Average heart rate moving
    - Average temperature
    """
    # special conditions for methods that raise Exceptions
    stats = _build_summary_statistics(obj)

    rows = {
        "Session": "Running: %s" % stats["start"].strftime("%d-%m-%Y %H:%M:%S"),
        "Total distance (meters)": stats["total_distance"],
        "Total ellapsed time": stats["ellapsed_time"],
        "Total moving time": stats["moving_time"],
        "Average speed (km/h)": stats["mean_speed"] * 3.6,
        "Average moving speed (km/h)": stats["mean_moving_speed"] * 3.6,
        "Average pace (per 1 km)": stats["mean_pace"],
        "Average pace moving (per 1 km)": stats["mean_moving_pace"],
        "Average cadence": stats["mean_cadence"],
        "Average moving cadence": stats["mean_moving_cadence"],
        "Average heart rate": stats["mean_heart_rate"],
        "Average moving heart rate": stats["mean_moving_heart_rate"],
        "Average temperature": stats["mean_temperature"],
    }

    series = pd.Series(
        rows,
        index=[
            "Session",
            "Total distance (meters)",
            "Total ellapsed time",
            "Total moving time",
            "Average speed (km/h)",
            "Average moving speed (km/h)",
            "Average pace (per 1 km)",
            "Average pace moving (per 1 km)",
            "Average cadence",
            "Average moving cadence",
            "Average heart rate",
            "Average moving heart rate",
            "Average temperature",
        ],
    )

    return series


def activity_summary(activity):
    """
    Returns the pandas Dataframe with the common basic statistics for the
    given activity.

    Parameters
    ----------
    activity:  runpandas.types.Activity. Runpandas Activity to be computed the statistics

    Returns
    -------
        pandas.Dataframe:  A pandas DataFrame containing the summary statistics, which
        inclues estimates of the total distance covered, the total duration,
        the time spent moving, and many others.

    """
    summary_statistics = _build_activity_statistics(activity)
    return summary_statistics.T


def session_summary(session):
    """
    Returns the a pandas Dataframe with the common basic statistics for the
    given activity.

    Parameters
    ----------
    session:  runpandas.types.Activity. Runpandas Activity with pandas.MultiIndex
    to be computed the statistics

    Returns
    -------
    pandas.Dataframe:  A pandas DataFrame containing the summary statistics
    across all th activities, which includes estimates of the total distance covered,
    the total duration, the time spent moving, and many others.

    """
    frames = []
    for index in session.index.unique(level="start"):
        df = session.xs(index, level=0)
        df.start = index
        frames.append(_build_session_statistics(df))

    session_summary = pd.concat(frames, axis=0, verify_integrity=True)
    session_summary.sort_index(inplace=True)
    return session_summary



"""
This is a module for extending pandas activities with session aggregation methods
"""

import pandas as pd
from runpandas.types import Activity
from runpandas.types import summary


@pd.api.extensions.register_dataframe_accessor("session")
class _SessionAcessor(object):
    """
    Dataframe Accessor to compute the aggregated metrics from a MultiIndex
    ``runpandas.types.Activity`` object.

    Raises
    ------
    AttributeError
        if it is not a ``runpandas.types.Activity`` instance or
        not have a ``pd.MultiIndex`` or not have the index levels
        of type ``pd.DateTimeIndex`` or ``pd.TimedeltaIndex``.

    """

    def __init__(self, session):
        self._validate(session)
        self._session = session

    @staticmethod
    def _validate(activity):
        assert isinstance(activity, Activity), "Activity is needed for this accessor"

        if not isinstance(activity.index, pd.MultiIndex) or not set(
            [*map(type, activity.index.levels)]
        ) == set([pd.TimedeltaIndex, pd.DatetimeIndex]):
            raise AttributeError(
                "An activity needs a DataFrame MultiIndex in \
                     (DateTimeIndex, TimedeltaIndex) format."
            )

    def summarize(self):
        """
        Summarize the session of activities by returning a Dataframe
        of the aggregated main statistics.

        Returns
        -------
        pandas.Dataframe: A Dataframe with the summarized statistics for the all the session.
        """
        return summary.session_summary(self._session)

    def count(self):
        """
        Returns the total number of activities in the session.

        Returns
        -------
        int: the total number of activities in the session.

        """
        return self._session.index.levshape[0]

    def distance(self, correct_distance=False, to_special_column=True, **kwargs):
        """
        Compute the distance per position and total distance acummulative
        over all the activities.

        Parameters
        ----------
        correct_distance: bool, optional
            It computes the distance corrected by the altitude. default is False.

        to_special_column: bool, optional
            It converts the distance calculated (`pandas.Series`) to special runpandas
            distance cummulative column (`runpandas.types.columns.DistancePerPosition`).
            Default is True.

        **kwargs: Keyword args to be passed to the `haversine` method

        Returns
        -------
            `runpandas.Activity`: The runpandas.Activity with the
             new columns dist and distpos computed.

        See Also
        --------
            runpandas.acessors.metrics.distance

        """
        for index in self._session.index.unique(level="start"):
            distpos = self._session.xs(index, level=0).compute.distance(
                correct_distance=correct_distance,
                to_special_column=to_special_column,
                **kwargs,
            )
            self._session.loc[
                pd.IndexSlice[index, distpos.index.tolist()], "distpos"
            ] = distpos.values
            dist = distpos.to_distance()
            self._session.loc[
                pd.IndexSlice[index, dist.index.tolist()], "dist"
            ] = dist.values
        return self._session

    def speed(self, from_distances=False, to_special_column=True, **kwargs):
        """
        Compute the speed in meters over all the activities.

        Parameters
        ----------
        from_distances: bool, optional
            Should the speeds be calculated from the distance recordings
            instead of taken from the speed recordings directly? Default is False.

        to_special_column: convert the distance calculated (`pandas.Series`)
            to special runpandas speed column (`runpandas.types.columns.Speed`).
            Default is True.

        **kwargs: Keyword args to be passed to the speed building method

        Returns
        -------
            `runpandas.Activity`: The runpandas.Activity with the new column speed computed.

        See Also
        --------
            runpandas.acessors.metrics.speed

        """
        for index in self._session.index.unique(level="start"):
            speed = self._session.xs(index, level=0).compute.speed(
                from_distances=from_distances,
                to_special_column=to_special_column,
                **kwargs,
            )
            self._session.loc[
                pd.IndexSlice[index, speed.index.tolist()], "speed"
            ] = speed.values
        return self._session

    def vertical_speed(self, to_special_column=True, **kwargs):
        """
        Compute the vertical climbing speed (VAM) in meters over all the activities.

        Parameters
        ----------
        from_distances: bool, optional
            Should the speeds be calculated from the distance recordings
            instead of taken from the speed recordings directly? Default is False.

        to_special_column: convert the distance calculated (`pandas.Series`)
            to special runpandas speed column (`runpandas.types.columns.Speed`).
            Default is True.

        **kwargs: Keyword args to be passed to the speed building method

        Returns
        -------
            `runpandas.Activity`: The runpandas.Activity with the new
             column vertical altitude speed computed.

        See Also
        --------
            runpandas.acessors.metrics.vertical_speed

        """
        for index in self._session.index.unique(level="start"):
            vam = self._session.xs(index, level=0).compute.vertical_speed(
                to_special_column=to_special_column, **kwargs
            )
            self._session.loc[
                pd.IndexSlice[index, vam.index.tolist()], "vam"
            ] = vam.values
        return self._session

    def gradient(self, to_special_column=True, **kwargs):
        """
        Compute the gradient ratio over all the activities.

        Parameters
        ----------
        to_special_column: convert the distance calculated (`pandas.Series`)
            to special runpandas Gradient column (`runpandas.types.columns.Gradient`).
            Default is True.

        **kwargs: Keyword args to be passed to the Gradient build method

        Returns
        -------
            `runpandas.Activity`: The runpandas.Activity with the new column gradient computed.

        See Also
        --------
            runpandas.acessors.metrics.gradient

        """
        for index in self._session.index.unique(level="start"):
            grad = self._session.xs(index, level=0).compute.gradient(
                to_special_column=to_special_column, **kwargs
            )
            self._session.loc[
                pd.IndexSlice[index, grad.index.tolist()], "grad"
            ] = grad.values
        return self._session

    def pace(self, to_special_column=True, **kwargs):
        """
        Compute the pace (the time that it takes to cover distances in your activities)
        over all the activities.

        Parameters
        ----------
        to_special_column: convert the pace calculated (`pandas.Series`)
            to special runpandas Pace column (`runpandas.types.columns.Pace`).
            Default is True.

        **kwargs: Keyword args to be passed to the Pace build method

        Returns
        -------
            `runpandas.Activity`: The runpandas.Activity with the new column pace computed.

        See Also
        --------
            runpandas.acessors.metrics.pace

        """
        for index in self._session.index.unique(level="start"):
            pace = self._session.xs(index, level=0).compute.pace(
                to_special_column=to_special_column, **kwargs
            )
            self._session.loc[
                pd.IndexSlice[index, pace.index.tolist()], "pace"
            ] = pace.values
        return self._session

    def heart_zone(self, bins, labels, **kwargs):
        """
        Compute the heart zone (with the training zone labels for each heart rate record)
        over all the activities.

        Parameters
        ----------
        bins: Left and right bounds for each zone.
                An example of valid bins are [0, 100, 140, 160, 999].

        labels: Specifies the labels for the zones.
                Must be the same length as the resulting zones.
                Example of valid labels is ["Z1", "Z2", "Z3", "Z4", "Z5"].

        **kwargs: Keyword args to be passed to the heart_zone build method

        Returns
        -------
            `runpandas.Activity`: The runpandas.Activity with the
                new column heart zone labels computed.

        See Also
        --------
            runpandas.acessors.metrics.heart_zone

        """
        for index in self._session.index.unique(level="start"):
            hr_zone = self._session.xs(index, level=0).compute.heart_zone(
                bins=bins, labels=labels, **kwargs
            )
            self._session.loc[
                pd.IndexSlice[index, hr_zone.index.tolist()], "hr_zone"
            ] = hr_zone.values
        return self._session

    def only_moving(self, threshold=0.8):
        """
        Impute the periods of inactivity, as such periods must be
        detected and ignored from the analysis for moving time, heart rate, speed, and
        power-based calculations for all the activities in the session.

        Parameters
        ----------
        threshold: float, default 0.8
            When the speed of a record drops below the threshold speed a 'false' event
            is set to the 'moving'column, and when the speed rises above the
            threshold speed a 'true' event is set to the 'moving' column.

        Returns
        -------
            `runpandas.Activity`: The runpandas.Activity with the column moving computed.

        See Also
        --------
            runpandas.acessors.moving._InactivityAssessor
        """
        for index in self._session.index.unique(level="start"):
            moving_series = self._session.xs(index, level=0).only_moving(
                threshold=threshold
            )["moving"]
            self._session.loc[
                pd.IndexSlice[index, moving_series.index.tolist()], "moving"
            ] = moving_series.values
        return self._session



"""Helpful utilities for runpandas modules.
"""
import os
import re
from xml.etree.cElementTree import iterparse
from functools import wraps
from runpandas import exceptions
from pandas import Series, Timedelta


def convert_pace_secmeters2minkms(seconds):
    """
    Convert pace from sec/meters to min/km.
    """
    pace_min = int((seconds * 1000) / 60)
    pace_sec = int(seconds * 1000 - (pace_min * 60))
    total_seconds = (pace_min * 60) + pace_sec
    return Timedelta(seconds=total_seconds)


def file_exists(fname):
    """Check if a file exists and is non-empty."""
    try:
        return fname and os.path.exists(fname) and os.path.getsize(fname) > 0
    except OSError:
        return False


def splitext_plus(fname):
    """Split on file extensions, allowing for zipped extensions."""
    base, ext = os.path.splitext(fname)
    if ext in [".gz", ".bz2", ".zip"]:
        base, ext2 = os.path.splitext(base)
        ext = ext2 + ext
    return base, ext.lower()


def is_valid(fname):
    """Check if it is a valid format for activity files.

    Parameters
    ----------
    fname : str
        Path to the file to be read.

    Returns
    -------
    It returns True if the it is a valid format for activities handling.
    """
    _, ext = os.path.splitext(fname)
    return ext.lower() in [".tcx", ".gpx", ".fit"]


def recursive_text_extract(node):
    ds = {}
    stack = []
    for child in node.iter():
        # print(sans_ns(child.tag), child.text, stack)
        if child.text is not None and child.text.strip():
            if sans_ns(child.tag) == "Value":
                tag = stack.pop()
            else:
                tag = child.tag
            ds[sans_ns(tag)] = child.text
        else:
            stack.append(child.tag)
    return ds


def sans_ns(tag):
    """Remove the namespace prefix from a tag."""
    return tag.split("}")[-1]


def get_nodes(file_path, node_names, *, with_root=False):
    """Parse XML document and iterate over specific nodes

    Parameters
    ----------
    filepath : str
        Path to the file to be read.
    node_names : str
        List of nodes to be extracted.
    with_root: boolean
        Default to False. If True returns the root node.

    Returns
    -------
    Used as a generator for yielding the nodes.
    """
    context = iter(iterparse(file_path, events=("start", "end")))
    event, root = next(context)

    if with_root:
        yield root

    for event, element in context:
        if event == "end" and sans_ns(element.tag) in node_names:
            yield element
            root.clear()


def camelcase_to_snakecase(string):
    """Converts the Camelcase string to snakecase string
    Example: ColumnName --> column_name
    Parameters
    ----------
    string : str
        String to be converted.

    Returns
    -------
    The string converted to lowercase.
    """
    string = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", string)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", string).lower()


def special_column(required_columns, name=None):
    """
    Decorator for certain methods of acessors that create special columns
    using the ``runpandas.types.MeasureSeries`` subtypes.

    Parameters
    ----------

    required_columns: tuple
        A tuple of column names.

    name: str, optional.
        The name for the returned Series object.
    """

    def real_decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            for column in required_columns:
                if column not in self._activity:
                    raise exceptions.RequiredColumnError(column)

            # If it's ok so construct the new Series.
            out = func(self, *args, **kwargs)
            if "to_special_column" in kwargs and kwargs["to_special_column"] is False:
                return Series(out, index=self._activity.index, name=name)
            return out

        return wrapper

    return real_decorator


class series_property:
    """A simple descriptor that emulates property, but returns a Series."""

    def __init__(self, fget):
        self.fget = fget

    def __get__(self, obj, objtype=None):
        return Series(self.fget(obj), index=obj.index)



"""Schema Models for loading example datasets"""

from typing import List
from enum import Enum
from pathlib import Path
from pydantic import BaseModel, validator


class MetricsEnum(Enum):
    latitude = "latitude"
    longitude = "longitude"
    speed = "speed"
    power = "power"
    elevation = "elevation"
    cadence = "cadence"
    heartrate = "heartrate"
    temperature = "temperature"
    distance = "distance"


class Sensor(BaseModel):
    name: str
    data_types: List[MetricsEnum]


class FileTypeEnum(str, Enum):
    FIT = "FIT"
    TCX = "TCX"
    GPX = "GPX"


class ActivityData(BaseModel):

    summary: str = None
    path: Path
    file_type: FileTypeEnum
    recording_device: str
    sensors: List[Sensor] = None
    included_data: List[MetricsEnum]
    laps: int = None
    sessions: int = None

    @validator("path")
    @classmethod
    def make_cached_path(cls, v):
        path = (
            "https://raw.githubusercontent.com/"
            "corriporai/runpandas-data/master/activities/{}"
        )
        return path.format(v)



"""
Test module for TCX reader base module
"""

import os
import pytest
from pandas import DataFrame, Timedelta, Timestamp
from runpandas import reader
from runpandas import types

pytestmark = pytest.mark.stable


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


@pytest.fixture
def pandas_activity(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    return reader._read_file(tcx_file, to_df=True)


@pytest.fixture
def runpandas_activity(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    return reader._read_file(tcx_file, to_df=False)


test_data = [
    (pytest.lazy_fixture("pandas_activity"), "heart_rate_bpm", 0, 62),
    (pytest.lazy_fixture("pandas_activity"), "heart_rate_bpm", -1, 180),
    (pytest.lazy_fixture("pandas_activity"), "altitude_meters", 0, 178.942626953),
    (pytest.lazy_fixture("pandas_activity"), "altitude_meters", -1, 166.4453125),
    (pytest.lazy_fixture("pandas_activity"), "latitude_degrees", 0, 35.951880198),
    (pytest.lazy_fixture("pandas_activity"), "longitude_degrees", 0, -79.0931872185),
    (pytest.lazy_fixture("pandas_activity"), "distance_meters", -1, 4686.31103516),
    (pytest.lazy_fixture("pandas_activity"), "distance_meters", 0, 0.0),
    (pytest.lazy_fixture("runpandas_activity"), "hr", 0, 62),
    (pytest.lazy_fixture("runpandas_activity"), "hr", -1, 180),
    (pytest.lazy_fixture("runpandas_activity"), "alt", 0, 178.942626953),
    (pytest.lazy_fixture("runpandas_activity"), "alt", -1, 166.4453125),
    (pytest.lazy_fixture("runpandas_activity"), "lat", 0, 35.951880198),
    (pytest.lazy_fixture("runpandas_activity"), "lon", 0, -79.0931872185),
    (pytest.lazy_fixture("runpandas_activity"), "dist", -1, 4686.31103516),
    (pytest.lazy_fixture("runpandas_activity"), "dist", 0, 0.0),
]


@pytest.mark.parametrize("activity,column,index,expected", test_data)
def test_tcx_values(activity, column, index, expected):
    assert activity[column].iloc[index] == pytest.approx(expected, 0.01)
    assert activity.index[-1] == Timedelta("0 days 00:33:11")

    if isinstance(activity, types.Activity):
        assert activity.start == Timestamp("2012-12-26 21:29:53+00:00")


def test_read_file_tcx_basic_dataframe(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    activity = reader._read_file(tcx_file, to_df=True)
    assert isinstance(activity, DataFrame)
    included_data = set(
        [
            "latitude_degrees",
            "longitude_degrees",
            "altitude_meters",
            "distance_meters",
            "heart_rate_bpm",
        ]
    )
    assert included_data <= set(activity.columns.to_list())
    assert activity.size == 1915


def test_read_file_tcx_basic_activity(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    activity = reader._read_file(tcx_file, to_df=False)
    assert isinstance(activity, types.Activity)
    included_data = set(["lat", "lon", "alt", "dist", "hr"])
    assert included_data <= set(activity.columns.to_list())
    assert activity.size == 1915



"""
Test module for Strava Client Authentication module
"""
import os
import json
import pytest
from unittest.mock import PropertyMock
from runpandas.io.strava._client import StravaClient, HTTPResponder


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


@pytest.fixture(scope="session")
def valid_token_file(tmpdir_factory):
    return tmpdir_factory.getbasetemp().join("token.json")


def test_stravaclient_with_enviroment_variables(dirpath):
    os.environ["STRAVA_CLIENT_SECRET"] = str("STRAVA_CLIENT_SECRET")
    os.environ["STRAVA_CLIENT_ID"] = str("STRAVA_ID")
    refresh_token = "REFRESHTOKEN"
    client = StravaClient(refresh_token=refresh_token)

    assert client.client_secret == "STRAVA_CLIENT_SECRET"
    assert client.client_id == "STRAVA_ID"
    assert client.token_file is None


def test_stravaclient_with_arguments(dirpath):
    os.environ["STRAVA_CLIENT_SECRET"] = str("STRAVA_CLIENT_SECRET")
    os.environ["STRAVA_CLIENT_ID"] = str("STRAVA_ID")
    refresh_token = "REFRESHTOKEN"
    client = StravaClient(
        refresh_token=refresh_token,
        client_id="STRAVA_ID",
        client_secret="STRAVA_CLIENT_SECRET",
    )

    assert client.client_secret == "STRAVA_CLIENT_SECRET"
    assert client.client_id == "STRAVA_ID"
    assert client.token_file is None

    # if token file is empty, so it can't write on it.
    assert client.save_token_to_file({"test": "test"}) is None


def test_stravaclient_rw_token_file(dirpath, valid_token_file):
    # create a token valid file
    file_handler = open(valid_token_file, "w")
    file_handler.write(
        '{"access_token": "2334444",  "refresh_token": "235555", \
             "expires_at": 1658341120.318284}'
    )
    file_handler.close()

    client = StravaClient(
        client_id="STRAVA_ID",
        client_secret="STRAVA_CLIENT_SECRET",
        token_file=valid_token_file,
    )

    assert client.access_token == "2334444"
    assert client.refresh_token == "235555"
    assert client.token_expires_at == 1658341120.318284

    assert client.client_secret == "STRAVA_CLIENT_SECRET"
    assert client.client_id == "STRAVA_ID"

    token = {
        "access_token": "222",
        "refresh_token": "111",
        "expires_at": 1658341120.318284,
    }
    client.save_token_to_file(token)
    with open(valid_token_file) as f:
        assert json.loads(f.read()) == json.loads(
            '{"access_token": "222",  "refresh_token": "111", "expires_at": 1658341120.318284}'
        )


def test_stravaclient_authenticate(dirpath, valid_token_file, mocker):
    mocker.patch("webbrowser.open")
    mocker.patch.object(HTTPResponder, "handle_request", return_value=None)
    mocker.patch.object(
        StravaClient,
        "exchange_code_for_token",
        return_value={
            "access_token": "1",
            "refresh_token": "1",
            "expires_at": 1658450893,
        },
    )
    mocker.patch.object(
        HTTPResponder, "access_token", new_callable=PropertyMock, return_value="123"
    )

    os.environ["STRAVA_CLIENT_SECRET"] = str("STRAVA_CLIENT_SECRET")
    os.environ["STRAVA_CLIENT_ID"] = str("STRAVA_ID")
    refresh_token = "REFRESHTOKEN"
    client = StravaClient(
        refresh_token=refresh_token,
        client_id="STRAVA_ID",
        client_secret="STRAVA_CLIENT_SECRET",
        token_file=valid_token_file,
    )
    client.authenticate_web()
    with open(valid_token_file) as f:
        returned = json.load(f)
        assert returned["access_token"] == "1"
        assert returned["refresh_token"] == "1"
        assert returned["expires_at"] == 1658450893



from runpandas.types import columns  # noqa
from runpandas.types.frame import Activity  # noqa
from runpandas.types.acessors import _InactivityAssessor  # noqa
from runpandas.types.acessors import MetricsAcessor  # noqa
from runpandas.types import summary  # noqa



"""
This is a module for extending pandas dataframes with the runpandas toolbox
"""

import pandas as pd


@pd.api.extensions.register_dataframe_accessor("only_moving")
class _InactivityAssessor(object):
    """
    Dataframe Accessor to impute the periods of inactivity, as such periods must be
    detected and ignored from the analysis for moving time, heart rate, speed, and
    power-based calculations.

    Parameters
    ----------
    threshold: float, default 0.8
        When the speed of a record drops below the threshold speed a 'false' event
        is set to the 'moving'column, and when the speed rises above the
        threshold speed a 'true' event is set to the 'moving' column.

    Raises
    ------
    AttributeError
        if at least one of the needed columns is missing.

    """

    # Speeds less than or equal to this value (in m/s) are considered to be stopped
    STOPPED_THRESHOLD = 0.8

    required_columns = ["speed"]

    def __init__(self, activity):
        self._validate(activity)
        self.activity = activity

    def __call__(self, threshold=STOPPED_THRESHOLD):
        """
        When the speed of a record drops below the threshold speed a 'false' event
        is set to the 'moving'column, and when the speed rises above the
        threshold speed a 'true' event is set to the 'moving' column.
        """
        # get the time difference between tracking positions and resample it to 1s
        # remove all instantaneous with speed below threashold
        self.activity["moving"] = self.activity["speed"] >= threshold
        return self.activity

    @staticmethod
    def _validate(activity):
        if not isinstance(activity.index, (pd.DatetimeIndex, pd.TimedeltaIndex)):
            raise AttributeError(
                "An activity needs a DataFrame Index in DateTimeIndex format"
            )

        if any(
            [c not in activity.columns for c in _InactivityAssessor.required_columns]
        ):
            raise AttributeError(
                "To compute the periods of inactivity, "
                + "it must have the properties [%s]."
                % (", ".join(_InactivityAssessor.required_columns))
            )



"""
Root pandas accessors for DataFrames. It exposes the ActiviyData as DataFrames

"""
import warnings

import numpy as np
import pandas as pd
import runpandas.reader
from pandas.core.frame import DataFrame
from runpandas.types import columns


class Activity(pd.DataFrame):
    """
    An Activity object is a pandas.DataFrame that provides useful
    methods and has specific columns with special functionalities.
    In addition to the standard DataFrame constructor arguments,
    Activity also accepts the following optional arguments:

    Parameters
    ----------
    cspecs : dict of series (optional)
        Column Specifications (cspecs) of the ``Activity``
    start : str, datetime (optional)
        Start timestamp of the ``Activity``
    """

    # properties to propagate
    _metadata = ["start"]

    def __init__(self, *args, **kwargs):
        cspecs = kwargs.pop("cspecs", None)
        start = kwargs.pop("start", None)

        super().__init__(*args, **kwargs)

        if cspecs is not None:
            self.set_specs(cspecs, inplace=True)
        if start is not None:
            self.start = start

    def set_specs(self, cspecs=None, inplace=False):
        """
        Set the Column Specifications (cspecs) of the ``Activity``.
        Set the specifed column on Activity.

        Parameters
        ----------
        cspecs : dict of series
        inplace : bool, default False
            If True, the cspecs of the Activity will be changed in place
            (while still returning the result) instead of making a copy of
            the Activity.
        """
        if not inplace:
            df = self.copy()
        else:
            df = self
        for old_key, column_cls in cspecs.items():
            try:
                old_column = df.pop(old_key)  # no default
            except KeyError:
                warnings.warn("The specified key %s not found." % old_key, UserWarning)
                continue

            new = column_cls(old_column)
            df[new.colname] = new
        if not inplace:
            return df

    @property
    def _constructor(self):
        return Activity

    def to_pandas(self):
        """
        Returns:
            Casted Pandas Dataframe from Activity
        """
        return DataFrame(self)

    def __finalize__(self, other, method=None, **kwargs):
        """Propagate metadata from other to self."""
        for name in self._metadata:
            object.__setattr__(self, name, getattr(other, name, None))
        return self

    def __getitem__(self, key):
        """
        If the result is a column containing only activity measures, return a
        MeasuredSeries. If it's a DataFrame with a 'Measures' column, return a
        Activity.
        """
        result = super(Activity, self).__getitem__(key)

        if isinstance(key, str) and key in columns.ColumnsRegistrator.REGISTRY:
            result.__class__ = columns.ColumnsRegistrator.REGISTRY[key]
        elif isinstance(result, pd.DataFrame):
            result.__class__ = Activity

        if isinstance(result, columns.Gradient):
            result._set_attrs(
                _rise=self["alt"].diff().values, _run=self["dist"].diff().values
            )

        return result

    @classmethod
    def from_file(cls, file_path, **kwargs):
        """
        Alternate constructor to create a ``Activity`` from a file.

        It is recommended to use :func:`runpandas.reader._read_file` instead.

        Parameters
        ----------
        file_path : str
            File path or file handle to read from. Depending on which kwargs
            are included, the content of filename may vary.
        kwargs : key-word arguments
            These arguments are passed to reader._read_file
        """
        return runpandas.reader._read_file(file_path)

    @property
    def ellapsed_time(self):
        """
        Returns:
            The duration of activity in `pandas.TimedeltaIndex` object.

        Raises:
            AttributeError if dataframe index is not an instance of TimedeltaIndex

        """
        if isinstance(self.index, pd.TimedeltaIndex):
            return self.index[-1]
        raise AttributeError("index is not TimedeltaIndex")

    @property
    def moving_time(self):
        """
        Returns:
            The moving time of activity in `pandas.TimedeltaIndex` object.
            It measures only the periods that you were active based on
            internal calculations.

        Raises:
            AttributeError if dataframe index is not an instance of TimedeltaIndex
            or the `moving` column is not found computed from
            `runpandas.acessors.moving.only_moving` acessor.

        """
        if not isinstance(self.index, pd.TimedeltaIndex):
            raise AttributeError("index is not TimedeltaIndex")
        if "moving" not in self.columns:
            raise AttributeError("moving column not found in activity.")

        total_time = (
            self.index.to_series().diff().fillna(self.index[0])
        ) / np.timedelta64(1, "s")
        merged_df = total_time.to_frame().join(self["moving"].to_frame())
        return pd.Timedelta(
            seconds=total_time.sum()
            - merged_df[~merged_df["moving"].astype(bool)]["time"].sum()
        )

    @property
    def distance(self):
        """
        Returns:
            The total distance of the activity in meters as a float number.
        """
        try:
            return self["dist"].max()
        except KeyError:
            return self["distpos"].sum()

    def mean_speed(self, only_moving=False, smoothing=True):
        """
        It calculates the average speed based on the speed trace.

        Parameters
        ----------
        only_moving : boolean, optional. It considers only the active moviment of the activity.
        Default is False.

        smoothing: boolean, optional. If True, it calculates average speed based on total
        distance divided by total time. Default is True.

        Returns:
        --------
            The average speed in meters/second for the activity.
        """
        if "speed" not in self.columns:
            raise AttributeError("speed column not found in activity.")

        if only_moving:
            total_time = self.moving_time
            activity = self[self["moving"]]
        else:
            total_time = self.ellapsed_time
            activity = self

        if smoothing:
            time_diff = (
                self.index.to_series().diff().fillna(self.index[0])
            ) / np.timedelta64(1, "s")
            total_distance = (activity["speed"] * time_diff).sum()
        else:
            total_distance = activity.distance

        return total_distance / total_time.total_seconds()

    def mean_pace(self, only_moving=False, smoothing=True):
        """
        It calculates the average pace based on the speed trace.

        Parameters
        ----------
        only_moving : boolean, optional. It considers only the active moviment of the activity.
        Default is False.

        smoothing: boolean, optional. If True, it calculates average pace based on total
        distance divided by total time. Default is True.

        Returns:
        --------
            The average pace in sec/m for the activity.
        """
        speed = self.mean_speed(only_moving, smoothing)
        return pd.Timedelta(seconds=1 / speed)

    def mean_heart_rate(self, only_moving=False):
        """
        It calculates the average heart rate based on the heart rate tracked
        by HeartRate Series column.

        Parameters
        ----------
        only_moving : boolean, optional. It considers only the active moviment of the activity.
        Default is False.

        Returns:
        --------
            The average heart zone in bpm for the activity.
        """
        if "hr" not in self.columns:
            raise AttributeError("heart rate column not found in activity.")

        if only_moving:
            activity = self[self["moving"]]
        else:
            activity = self

        return activity["hr"].mean()

    def mean_cadence(self, only_moving=False):
        """
        It calculates the average cadence based on the cadence tracked
        by Cadence Series column.

        Parameters
        ----------
        only_moving : boolean, optional. It considers only the active moviment of the activity.
        Default is False.

        Returns:
        --------
            The average cadence in rotation per minute (rpm) for the activity.
        """
        if "cad" not in self.columns:
            raise AttributeError("Cadence column not found in activity.")

        if only_moving:
            activity = self[self["moving"]]
        else:
            activity = self

        return activity["cad"].mean()

    def summary(self):
        """
        Display the common basic statistics for the activity.

        Returns:
        --------
            pandas.Dataframe:  A pandas DataFrame containing the summary statistics, which
            inclues estimates of the total distance covered, the total duration,
            the time spent moving, and many others.

        """
        return runpandas.types.summary.activity_summary(self)



"""
Test module for utilities from the datasets package
"""

import os
import shutil
import pytest
from runpandas.datasets.utils import (
    _get_config_data,
    _get_cache_path,
    _get_activity_index,
    activity_examples,
)
from runpandas.datasets.schema import ActivityData, FileTypeEnum
from runpandas import read_file
from runpandas.types import Activity

pytestmark = pytest.mark.stable


@pytest.fixture
def dirpath(datapath):
    return datapath("datasets")


def test_config_data(dirpath):
    # test default config data
    config = _get_config_data()
    assert "runpandas_data" in config["path"]["root"]

    # test custom config data
    test_config_file = os.path.join(dirpath, "config.test.yaml")

    config = _get_config_data(test_config_file)
    assert "test_runpandas_data" in config["path"]["root"]


def test_get_cache_path(dirpath):
    # first create
    test_config_file = os.path.join(dirpath, "config.test.yaml")
    directory = _get_cache_path(test_config_file)
    assert os.path.exists(directory)

    # test after created
    directory = _get_cache_path(test_config_file)
    assert os.path.exists(directory)
    shutil.rmtree(directory)


def test_get_activity_index():
    index = _get_activity_index()
    assert len(index) > 0
    assert type(index[0]) is ActivityData
    assert "polarm400.tcx" in [os.path.basename(item.path) for item in index]


def test_activity_examples(dirpath):
    # test filtered activities examples
    tcx_examples = activity_examples(file_type=FileTypeEnum.TCX)
    assert "polarm400.tcx" in [os.path.basename(item.path) for item in tcx_examples]

    # test empty actitivies examples
    empty_examples = activity_examples(file_type="PWX")
    assert len(list(empty_examples)) == 0

    # test not found file
    with pytest.raises(ValueError):
        _ = activity_examples(path="test.tcx")

    # test not cached file
    test_config_file = os.path.join(dirpath, "config.test.yaml")
    example_activity = activity_examples(path="polarm400.tcx", config=test_config_file)
    assert "polarm400.tcx" in example_activity.path
    assert os.path.exists(example_activity.path)
    assert type(read_file(example_activity.path)) is Activity

    # test cached file
    example_activity = activity_examples(path="polarm400.tcx", config=test_config_file)
    assert "polarm400.tcx" in os.path.basename(example_activity.path)
    assert os.path.exists(example_activity.path)
    assert type(read_file(example_activity.path)) is Activity

    # test fit file
    example_activity = activity_examples(
        path="Garmin_Fenix_6S_Pro-Running.fit", config=test_config_file
    )
    assert "Garmin_Fenix_6S_Pro-Running.fit" in os.path.basename(example_activity.path)
    assert os.path.exists(example_activity.path)
    assert type(read_file(example_activity.path)) is Activity

    directory = _get_cache_path(test_config_file)
    shutil.rmtree(directory)



"""
Tests for XML tools in utils module
"""

import pytest
import io
from runpandas import _utils as utils

pytestmark = pytest.mark.stable


data = """<?xml version="1.0" encoding="UTF-8" ?>
<TrainingCenterDatabase>
<Activities>
<Activity Sport="Running">
<Id>2020-06-28T09:39:21Z</Id>
<Lap StartTime="2020-06-28T09:39:21Z">
<TotalTimeSeconds>552</TotalTimeSeconds>
<DistanceMeters>1007.838196</DistanceMeters>
<Calories>99</Calories>
<AverageHeartRateBpm>
<Value>128</Value>
</AverageHeartRateBpm>
<MaximumHeartRateBpm>
<Value>165</Value>
</MaximumHeartRateBpm>
<Intensity>Active</Intensity>
<TriggerMethod>Manual</TriggerMethod>
<Track>
<Trackpoint>
<Time>2020-06-28T09:39:24Z</Time>
<Position>
<LatitudeDegrees>-8.367331</LatitudeDegrees>
<LongitudeDegrees>-36.576199</LongitudeDegrees>
</Position>
<AltitudeMeters>679.053528</AltitudeMeters>
<DistanceMeters>5320.761719</DistanceMeters>
<HeartRateBpm>
<Value>62</Value>
</HeartRateBpm>
<Extensions>
<TPX xmlns="http://www.garmin.com/xmlschemas/ActivityExtension/v2">
<Speed>0.000000</Speed>
</TPX>
</Extensions>
</Trackpoint>
<Trackpoint>
<Time>2020-06-28T09:39:33Z</Time>
<Position>
<LatitudeDegrees>-8.364550</LatitudeDegrees>
<LongitudeDegrees>-36.578003</LongitudeDegrees>
</Position>
<AltitudeMeters>668.866272</AltitudeMeters>
<DistanceMeters>12.899824</DistanceMeters>
<HeartRateBpm>
<Value>60</Value>
</HeartRateBpm>
<Extensions>
<TPX xmlns="http://www.garmin.com/xmlschemas/ActivityExtension/v2">
<Speed>0.000000</Speed>
</TPX>
</Extensions>
</Trackpoint>
</Track>
</Lap>
</Activity>
</Activities>
</TrainingCenterDatabase>"""


def test_gen_nodes():
    faketcx = io.StringIO(data)

    trackpoints = [
        utils.recursive_text_extract(trkpt)
        for trkpt in utils.get_nodes(faketcx, ("Trackpoint",))
    ]

    assert len(trackpoints) == 2
    t1, t2 = trackpoints
    assert len(t1) == 7 and len(t2) == 7
    assert t1["Speed"] == "0.000000" and t2["Speed"] == "0.000000"
    assert t1["Time"] == "2020-06-28T09:39:24Z"
    assert t2["Time"] == "2020-06-28T09:39:33Z"
    assert t1["HeartRateBpm"] == "62" and t2["HeartRateBpm"] == "60"



#!/usr/bin/env/python
"""Installation script

"""

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from setuptools import setup, find_packages

import versioneer

LONG_DESCRIPTION = """RunPandas is a project to add support for data collected
by GPS-enabled tracking devices, heart rate monitors data to `pandas`_ objects.

RunPandas is a Python package that provides infrastructure for importing
tracking data from such devices, enabling statistical and visual analysis
for running enthusiasts and lovers. Its goal is to fill the gap between
the routine collection of data and their manual analyses in Pandas and Python.

.. _pandas: http://pandas.pydata.org
"""


INSTALL_REQUIRES = []
TEST_REQUIRES = []
with open("./requirements.txt") as f:
    INSTALL_REQUIRES = f.read().splitlines()
with open("./requirements-dev.txt") as f:
    TEST_REQUIRES = f.read().splitlines()

# get all data dirs in the datasets module
data_files = []

setup(
    name="runpandas",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Running Tracking Analysis",
    license="BSD",
    author="RunPandas contributors",
    author_email="marcel@caraciolo.com.br",
    url="https://github.com/corriporai/runpandas",
    long_description=LONG_DESCRIPTION,
    python_requires=">=3.6",
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    packages=find_packages(exclude=["tests*"]),
    zip_safe=False,
)



"""
Custom exceptions for runpandas
"""


class InvalidFileError(Exception):
    def __init__(self, msg):
        message = "It doesn't like a valid  %s file!" % (msg)
        super().__init__(message)


class RequiredColumnError(Exception):
    def __init__(self, column, cls=None):
        if cls is None:
            message = "{!r} column not found".format(column)
        else:
            message = "{!r} column should be of type {!s}".format(column, cls)
        super().__init__(message)
